import numpy as np
from numba import jit, float64, prange, objmode, typeof
import math
from numba.typed import List
import ctypes
from scipy.stats import norm
from numba.extending import get_cython_function_address

lgamma = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)(
    get_cython_function_address('scipy.special.cython_special', 'gammaln'))
digamma = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)(
    get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1psi'))

@jit(nopython=True,fastmath=True)
def binarize(input):
    """Computes binary adjacency matrix from weighted matrix.
    """
    dim = len(input)
    byn = np.zeros(dim)
    for i in range(dim):
        if input[i]> 0:
            byn[i] = 1.
    return byn 

@jit(nopython=True,fastmath=True)
def ll_logit(params,Wij,selection_variables,fixed_selection_params = np.array([])):
    """Computes opposite of the  log-likelihood for Logit model
    """
    Aij = binarize(Wij)
    n_params = len(params)
    params = np.concatenate((params,fixed_selection_params))
    Gi = np.exp(selection_variables @ params)
    
    n_countries = int(np.sqrt(len(Wij)))
    ll = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                aij = Aij[ij]
                Gij = Gi[ij]

                ll += aij*np.log(Gij) - np.log(1+Gij)
    
    return - ll

@jit(nopython=True,fastmath=True)
def jac_logit(params,Wij,selection_variables,fixed_selection_params = np.array([])):
    """Computes opposite of the  jacobian for Logit model"""
    
    Aij = binarize(Wij)
    n_params = len(params)
    params = np.concatenate((params,fixed_selection_params))
    Gi = np.exp(selection_variables @ params)
    
    G_matrix = np.zeros(len(Aij))
    n_countries = int(np.sqrt(len(Wij)))
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                ai = Aij[ij]
                G_matrix[ij] = ai - Gi[ij]/(1+Gi[ij])
    
    jac = (selection_variables.T @ G_matrix)[:n_params]

    return - jac
    
@jit(nopython=True,fastmath=True)
def ll_BCM(params,Wij):
    """Computes opposite of the  log-likelihood for Undirected Binary Configuration Model """
    Aij = binarize(Wij)
    n_countries = int(np.sqrt(len(Aij)))

    ll = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                aij = Aij[ij]
                xij = np.exp(-params[i]-params[j])

                ll += aij*np.log(xij) - np.log(1+xij)
    return - ll

@jit(nopython=True,fastmath=True)
def jac_BCM(params,Wij):
    """Computes opposite of the  jacobian for Undirected Binary Configuration Model """
    
    Aij = binarize(Wij)
    
    n_countries = int(np.sqrt(len(Aij)))
    jac = np.zeros(len(params))
    for i in range(n_countries):
        
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                aij = Aij[ij]
                xij = np.exp(-params[i]-params[j])

                jac[i]+= xij/(1+xij) - aij
    return - jac

@jit(nopython=True,fastmath=True)
def hess_BCM(params,Wij):
    """Computes opposite of the  hessian for Undirected Binary Configuration Model """
    
    Aij = binarize(Wij)
    
    n_countries = int(np.sqrt(len(Aij)))
    hess = np.zeros((len(params),len(params)))
    for i in range(n_countries):
        
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                aij = Aij[ij]
                xij = np.exp(-params[i]-params[j])
                pij = xij/(1.+xij)
                hess[i,i] += - pij*(1.-pij)
                hess[i,j] = -pij*(1.-pij)
                
                
    return - hess

@jit(nopython=True,fastmath=True)
def ll_DBCM(params,Wij):
    """Computes opposite of the  log-likelihood for Directed Binary Configuration Model"""
    
    Aij = binarize(Wij)
    n_countries = int(np.sqrt(len(Aij)))
    params_out = params[:n_countries]
    params_in = params[n_countries:]
    ll = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                aij = Aij[ij]
                xij = np.exp(-params_out[i]-params_in[j])

                ll += aij*np.log(xij) - np.log(1+xij)
    return - ll

@jit(nopython=True,fastmath=True)
def jac_DBCM(params,Wij):
    """Computes opposite of the  jacobian for Directed Binary Configuration Model"""
    
    Aij = binarize(Wij)
    n_countries = int(np.sqrt(len(Aij)))
    
    params_out = params[:n_countries]
    params_in = params[n_countries:]
    
    jac = np.zeros(len(params))
    for i in range(n_countries):
        
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                ji = j*n_countries + i
                aij = Aij[ij]
                aji = Aij[ji]
                
                xij = np.exp(-params_out[i]-params_in[j])
                xji = np.exp(-params_out[j]-params_in[i])
                jac[i]+= xij/(1+xij) - aij
                jac[i+n_countries]+= xji/(1.+xji) - aji
    return - jac

@jit(nopython=True,fastmath=True)
def hess_DBCM(params,Wij):
    """Computes opposite of the  hessian for Directed Binary Configuration Model"""
    
    Aij = binarize(Wij)
    
    n_countries = int(np.sqrt(len(Aij)))
    hess = np.zeros((len(params),len(params)))
    params_out = params[:n_countries]
    params_in = params[n_countries:]
    
    for i in range(n_countries):
        
        for j in range(n_countries):
            if j==i:
                for k in range(n_countries):
                    if k!=i:
                        xik = np.exp(-params_out[i]-params_in[k])
                        xki = np.exp(-params_in[i]-params_out[k])
            
                        pij = xik/(1.+xik)
                        pji = xki/(1.+xki)
                         
                        hess[i,j] += - pij*(1.-pij)
                        hess[i+n_countries,j+n_countries] += - pji*(1.-pji) 
            elif j!=i:
                xij = np.exp(-params_out[i]-params_in[j])
                xji = np.exp(-params_out[j]-params_in[i])
                pij = xij/(1.+xij)
                pji = xji/(1.+xji)
                
                hess[i,j+n_countries] = - pij*(1.-pij)
                hess[i+n_countries,j] = -pji*(1.-pji)
                
                
    return - hess

@jit(nopython=True,fastmath=True)
def ll_CGeom(params,Wij,exogenous_variables):
    "Computes the opposite of the  log-likelihood for the weighted part of the Conditional Geometric model"
    params_exog = params[:-1]
    z_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                denominator = 1 + zi - z_0*zi
                if ai == 1:
                    ll += (yi-ai)*np.log(z_0*zi/(1+zi)) + ai*np.log(denominator/(1+zi))
                
    return - ll 

    


@jit(nopython=True,fastmath=True)
def jac_CGeom(params,Wij,exogenous_variables):
    "Computes the opposite of the  jacobian for the weighted part of the Conditional Geometric model"
    
    params_exog = params[:-1]
    
    z_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    jac = np.zeros(len(params))
    n_obs = len(Aij)
    mu_matrix = np.zeros(n_obs)
    adder = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                if ai == 1:
                    
                    zi = mu[ij]
                    denominator = (1+zi-z_0*zi)
                    yi_hat = (1+zi)/denominator

                    mu_matrix[ij] = yi/(1+zi) - ai/denominator
                    adder += - yi + ai*yi_hat
        
        
    jac[:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder
    
    return - jac


@jit(nopython=True,fastmath=True)
def hess_CGeom(params,Wij,exogenous_variables):
    "Computes the opposite of the  hessian for the weighted part of the Conditional Geometric model"
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[:-1]
    beta_0 = params[-1]
    
    z_0 = np.exp(-beta_0)
    x1_beta1 = exogenous_variables @ params_exog
    mu = np.exp(x1_beta1)
    
    Aij = binarize(Wij)
    n_obs = len(Aij)
        
    
    hess = np.zeros((len(params),len(params)))

    W_beta_1 = np.zeros((n_obs,n_obs))
    W_beta_0 = np.zeros(n_obs)
    W_beta_mixed = np.zeros(n_obs)

    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
    
                ai = Aij[ij]
                yi = Wij[ij]
                zi = mu[ij]
                denominator = 1 + zi - z_0*zi
                yi_hat = (1+zi)/denominator
                
                W_beta_0[ij] = - ai*zi*z_0*(1+zi)/denominator**2

                W_beta_1[ij,ij] = - zi*(yi-ai*yi_hat)/(1+zi)**2 + ai*z_0/((denominator**2)*(1+zi))
                W_beta_mixed[ij] = - ai*z_0*zi/denominator**2

    
    hess[:n_exog_variables,:n_exog_variables] = exogenous_variables.T @ W_beta_1 @ exogenous_variables
    hess[-1,:n_exog_variables] = (exogenous_variables.T @ W_beta_mixed)
    hess[:n_exog_variables,-1] = (exogenous_variables.T @ W_beta_mixed)
    hess[-1,-1] = W_beta_0.sum()

    return - hess


@jit(nopython=True,fastmath=True)
def ll_CExp(params,Wij,exogenous_variables):
    "Computes the  opposite of the  log-likelihood for the weighted part of the Conditional Exponential model"
    
    params_exog = params[:-1]
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                if ai == 1:
                    
                    zi = mu[ij]
                    ll += ai*np.log(1.+beta_0*zi) - ai* np.log(zi) - yi*(beta_0+1./zi)
                    
    return - ll 

    


@jit(nopython=True,fastmath=True)
def jac_CExp(params,Wij,exogenous_variables):
    "Computes the opposite of the  jacobian for the weighted part of the Conditional Exponential model"
    
    params_exog = params[:-1]
    
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    jac = np.zeros(len(params))
    n_obs = len(Aij)
    mu_matrix = np.zeros(n_obs)
    adder = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                if ai == 1:
                    
                    zi = mu[ij]
                    yi_hat = zi/(1+beta_0*zi)

                    mu_matrix[ij] = (yi-ai*yi_hat)/zi
                    adder += - yi + ai*yi_hat
                    
        
    jac[:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder
    
    return - jac


    


@jit(nopython=True,fastmath=True)
def hess_CExp(params,Wij,exogenous_variables):
    "Computes the opposite of the  hessian for the weighted part of the Conditional Exponential model"
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[:-1]
    beta_0 = params[-1]
    
    
    x1_beta1 = exogenous_variables @ params_exog
    mu = np.exp(x1_beta1)
    
    Aij = binarize(Wij)
    n_obs = len(Aij)
        
    
    hess = np.zeros((len(params),len(params)))

    W_beta_1 = np.zeros((n_obs,n_obs))
    W_beta_mixed = np.zeros(n_obs)

    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
    
                ai = Aij[ij]
                yi = Wij[ij]
                zi = mu[ij]
                denominator = 1 + beta_0*zi
                
                hess[-1,-1] += - ai*zi**2/denominator**2

                W_beta_1[ij,ij] = - yi/zi + ai*beta_0*zi/denominator**2
                W_beta_mixed[ij] = - ai*zi/denominator**2

    
    hess[:n_exog_variables,:n_exog_variables] = exogenous_variables.T @ W_beta_1 @ exogenous_variables
    hess[-1,:n_exog_variables] = (exogenous_variables.T @ W_beta_mixed)
    hess[:n_exog_variables,-1] = (exogenous_variables.T @ W_beta_mixed)
    
    return - hess



@jit(nopython=True,fastmath=True)
def ll_CPareto(params_exog,Wij,exogenous_variables):
    "Computes the opposite of the  log-likelihood for the weighted part of the Conditional Pareto model"
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_countries = int(np.sqrt(len(Aij)))
    
    positive_weights = List()
    for i in range(len(Wij)):
        wij = Wij[i]
        if wij > 0:
            positive_weights.append(wij)
    w_min = min(positive_weights)
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                if ai == 1:
                    ll += (np.log(1+zi) - np.log(zi) + (1.+1./zi)*np.log(w_min) - (2.+1./zi) * np.log(yi))
                
                
    return - ll 



@jit(nopython=True,fastmath=True)
def jac_CPareto(params_exog,Wij,exogenous_variables):
    "Computes the opposite of the jacobian for the weighted part of the Conditional Pareto model"
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    jac = np.zeros(len(params_exog))
    n_obs = len(Aij)
    mu_matrix = np.zeros(n_obs)
    n_countries = int(np.sqrt(len(Aij)))
    positive_weights= List()
    for ij in range(len(Wij)):
        weight = Wij[ij]
        if weight>0:
            positive_weights.append(weight)
    w_min = min(positive_weights)
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                lnyi_hat = np.log(w_min) + zi/(1.+zi)
                if ai == 1:
                    mu_matrix[ij] = (np.log(yi)-ai*lnyi_hat)/zi
                
                
        
    jac = exogenous_variables.T @ mu_matrix
    
    return - jac



@jit(nopython=True,fastmath=True)
def ll_CGamma(params,Wij,exogenous_variables):
    "Computes the opposite of the  log-likelihood for the weighted part of the Conditional Gamma model"
    
    params_exog = params[:-2]
    beta_0 = params[-2]
    phi = params[-1]
    x1_beta1 = exogenous_variables @ params_exog
    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                if ai == 1:
                    
                    zi = mu[ij]
                    ll += -np.log(yi) + phi*np.log(yi) + phi*np.log(beta_0 + 1./zi) - lgamma(phi) - (beta_0 + 1./zi)*yi
    return - ll 

@jit(nopython=True,fastmath=True)
def jac_CGamma(params,Wij,exogenous_variables):
    "Computes the opposite of the  jacobian for the weighted part of the Conditional Gamma model"
    
    params_exog = params[:-2]
    beta_0 = params[-2]
    phi = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    jac = np.zeros(len(params))
    n_obs = len(Aij)
    mu_matrix = np.zeros(n_obs)
    d_phi = 0
    d_beta_0 = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                if ai == 1:
                    
                    zi = mu[ij]
                    yi_hat = phi*zi/(1.+beta_0*zi) 
                    lnyi_hat = - np.log(beta_0 + 1./zi) + digamma(phi)
                    
                    mu_matrix[ij] = (yi - ai*yi_hat)/zi
                    d_phi += np.log(yi) - lnyi_hat
                    d_beta_0  += -yi + yi_hat 
                
        
    jac[:-2] = exogenous_variables.T @ mu_matrix
    jac[-2] = d_beta_0
    jac[-1] = d_phi
    
    return - jac





@jit(nopython=True,fastmath=True)
def ll_CLognormal(params,Wij,exogenous_variables):
    "Computes the opposite of the  log-likelihood for the weighted part of the Conditional Lognormal model"
    
    params_exog = params[:-1]
    gamma_0 = params[-1]
    mu = exogenous_variables @ params_exog   
    
    Aij = binarize(Wij)
    ll = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                ai = Aij[ij]
                if ai == 1:
                    
                    yi = Wij[ij]
                
                    log_yi = np.log(yi)
                    lnzi = mu[ij]

                    xi_ij = 1.-lnzi
                    ll+= - xi_ij**2/(4.*gamma_0) - xi_ij*log_yi - gamma_0*log_yi**2
                    ll+= -.5*np.log(math.pi/gamma_0) - (1.-2.*xi_ij)/(4.*gamma_0)

    return - ll 



@jit(nopython=True,fastmath=True)
def jac_CLognormal(params,Wij,exogenous_variables):
    "Computes the opposite of the  jacobian for the weighted part of the Conditional Lognormal model"
    
    params_exog = params[:-1]
    
    gamma_0 = params[-1]
    mu = exogenous_variables @ params_exog 
    
    Aij = binarize(Wij)
    jac = np.zeros(len(params))
    n_obs = len(Aij)
    mu_matrix = np.zeros(n_obs)
    adder = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                if ai == 1:
                    
                    lnzi = mu[ij]
                    log_yi = np.log(yi)
                    log_yi_hat = lnzi/(2.*gamma_0)
                    log_yi2_hat = 1./(2.*gamma_0) + log_yi_hat**2 
                    
                    mu_matrix[ij] = log_yi - log_yi_hat
                    adder += log_yi2_hat - log_yi**2.
                
    jac[:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder
    
    return - jac

@jit(nopython=True,fastmath=True)
def alpha_estimate_CLognormal(params_exog,Wij,exogenous_variables):
    """Computes an approximate estimate for the last parameter of the Lognormal model given the others."""
    mu = exogenous_variables @ params_exog
    Aij = binarize(Wij)
    n_countries = int(np.sqrt(len(Aij)))
    count = 0
    rhs = 0
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                if ai == 1:
                    count += 1
                    log_yi = np.log(yi)
                    rhs += (log_yi - zi)**2
                    
    sigma2 = rhs/(count)
    alpha = 1./(2.*sigma2)                
    return alpha







import numpy as np
from numba import jit,float64
from scipy import optimize as opt

from . import ll_Functions_conditional as lfC
from . import ll_Functions_integrated as lfI
from . import ll_Functions_econometrics as lfE

# @jit(float64[:](float64[:]),nopython=True,fastmath=True)
@jit(nopython=True)
def binarize(input):
    """Computes binary adjacency matrix from weighted adjacency matrix."""
    dim = len(input)
    byn = np.zeros(dim)
    for i in range(dim):
        if input[i]> 0:
            byn[i] = 1.
    return byn 

def solver(model,Wij,selection_variables,exogenous_variables,fixed_selection_params=np.array([]),tol=1e-5,
           use_guess= np.array([]), verbose=False, print_steps= 1, maxiter=20 ):  
    """Solves chosen model using scipy.optimize and scipy.least_squares routines.
    
    :param model: chosen model
    :type model: str
    :param Wij:  weighted adjacency matrix
    :type Wij: np,ndarray
    :param selection_variables: regressor matrix for the topological optimization of Zero-Inflated and L-constrained Conditional Models.
    :param exogenous_variables: regressor matrix for the weighted optimization
    :param fixed_selection_variables: fixed parameters for the topological optimization of Zero-Inflated and L-constrained Conditional Models. Defaults to np.array([1.00000,1.0000,0.0000]).
    :param tol: tolerance for optimization. Defaults to 1e-5.
    :param use_guess: optional starter guess for the optimization process. Defaults to np.array([]).
    :param verbose: True if you want to print iteration values of infinite norm. Defaults to False.
    :param print_steps: If verbose is True, you decide after how many steps you print on the screen. Defaults to 1.
    :param maxiter: Maxiter for optimization process. Defaults to 10.
    
    :type selection_variables: np.ndarray
    :type exogenous_variables: np.ndarray
    :type fixed_selection_variables: np.ndarray
    :type tol: float, optional
    :type use_guess: np.ndarray, optional
    :type verbose: bool, optional
    :type print_steps: int, optional
    :type maxiter: int, optional
    
    :raise TypeError: If chosen model is not correctly written or not implemented. See self.implemented_models to see available models.

    :return: estimated model solution
    :rtype: np.ndarray
    """
    
    if model == "Logit":
        n_select_variables = selection_variables.shape[1] - len(fixed_selection_params)
        if len(use_guess) == 0:
            sol = np.ones(n_select_variables)
        else:
            sol = use_guess    
        norm = 10000
        step = 0
        
        while norm > tol and step < maxiter:    
            
            sol = opt.least_squares(fun=lfC.jac_logit,x0 = sol, verbose=0,args=(Wij,selection_variables,fixed_selection_params)).x
            
            ll_lgt = - lfC.ll_logit(sol,Wij,selection_variables,fixed_selection_params)
            jac_lgt = - lfC.jac_logit(sol,Wij,selection_variables,fixed_selection_params)
            norm =np.linalg.norm(jac_lgt,ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps==0:
                    print('iteration:', step, 'norm:', norm, 'll:',ll_lgt)
        
        return sol
        
    elif model == "UBCM":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            sol = np.random.uniform(-10,10,n_countries)
        else:
            sol = use_guess
            
        norm = 1000
        step = 0
        
        while norm > tol and step < maxiter:
            sol =  opt.least_squares(fun=lfC.jac_BCM,x0 = sol, args=(Wij,)).x
            ll_bcm = - lfC.ll_BCM(sol,Wij)
            jac_bcm = lfC.jac_BCM(sol,Wij)
            norm = np.linalg.norm(jac_bcm,ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps==0:
                    print('iteration:',step,'norm:',norm,'ll:',ll_bcm)
        
        return sol
    
    elif model == "DBCM":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            sol = np.random.uniform(-10,10,2*n_countries)
        else:
            sol = use_guess
        norm = 1000
        step = 0
        
        while norm > tol and step < maxiter:
        
            sol = opt.least_squares(fun=lfC.jac_DBCM,x0 = sol, args=(Wij,)).x
            ll_bcm = -lfC.ll_DBCM(sol,Wij)
            jac_bcm = lfC.jac_DBCM(sol,Wij)
            norm = np.linalg.norm(jac_bcm,ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps ==0:
                    print('iteration:', step,'norm:',norm, 'll:',ll_bcm)
                    
        return sol
    
    elif model == "POIS":
        if len(use_guess) ==0:

            beta_0 = lfE.guess_constant_parameter(Wij,exogenous_variables)
            # print("beta_0: ",- beta_0)
            guess_aux = np.zeros(exogenous_variables.shape[1]-1)
            guess_UPOIS = np.array([- beta_0])
            sol = np.concatenate((guess_UPOIS,guess_aux))
            
        else:
            sol = use_guess
        
        norm = 1000
        step = 0
        
        while norm > tol and step < maxiter:
        
            sol = opt.minimize(fun=lfE.ll_POIS,x0=sol,method='Nelder-Mead',args=(Wij,exogenous_variables)).x
            sol = opt.least_squares(fun=lfE.jac_POIS,x0=sol,verbose=0,args=(Wij,exogenous_variables)).x
            ll_POIS = - lfE.ll_POIS(sol,Wij,exogenous_variables)
            jac_POIS = - lfE.jac_POIS(sol,Wij,exogenous_variables)
            norm = np.linalg.norm(jac_POIS,ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps ==0:
                    print('iteration:', step,'norm:',norm,'ll:',ll_POIS)
                    
            
        return sol
    
    elif model == "NB2":
        if len(use_guess) == 0:
            sol = np.ones(exogenous_variables.shape[1]+1)
        
        else:
            sol = use_guess
        
        
        norm = 1000
        step = 0
        
        while norm > tol and step < maxiter:
            
            sol = opt.minimize(fun=lfE.ll_NB2,x0=sol,method='Nelder-Mead',args=(Wij,exogenous_variables)).x
            sol = opt.least_squares(fun=lfE.jac_NB2,x0=sol,verbose=0,args=(Wij,exogenous_variables)).x
            ll_NB2 = - lfE.ll_NB2(sol,Wij,exogenous_variables)
            jac_NB2 = lfE.jac_NB2(sol,Wij,exogenous_variables)
            norm = np.linalg.norm(jac_NB2,ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm:',norm,'ll:',ll_NB2)
        
        return sol    
    
    
    elif model == "ZIP":
        n_select_variables = selection_variables.shape[1] - len(fixed_selection_params)
        if len(use_guess) == 0:
            sol = np.random.random(n_select_variables + exogenous_variables.shape[1]) 
        else:
            sol = use_guess
        norm = 1000
        step = 0
        
        while norm > tol and step < maxiter:
            # try:
            #     sol = opt.minimize(fun = lfE.ll_ZIP,jac=lfE.jac_ZIP, x0 = sol, method="BFGS",
            #                                         args=(Wij,selection_variables,exogenous_variables,fixed_selection_params)).x
            # except ZeroDivisionError:
            #     pass
            
            try:
                sol = opt.least_squares(fun = lfE.jac_ZIP, x0 = sol, verbose=0, 
                                                    args=(Wij,selection_variables,exogenous_variables,fixed_selection_params)).x
            except ZeroDivisionError:
                pass
            ll_ZIP = -lfE.ll_ZIP(sol,Wij,selection_variables,exogenous_variables,fixed_selection_params)
            jac = lfE.jac_ZIP(sol,Wij,selection_variables,exogenous_variables,fixed_selection_params)
            norm = np.linalg.norm(jac,ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm:',norm,'ll:',ll_ZIP)
                    
        return sol
    
    elif model == "ZINB":
        n_select_variables = selection_variables.shape[1] - len(fixed_selection_params)
        n_exog_variables = exogenous_variables.shape[1]
        if len(use_guess)==0:
            sol = np.ones(n_select_variables+n_exog_variables+1)
        else:
            sol = use_guess    
        
        step = 0
        norm = 10000
        while norm > tol and step < maxiter:
            try:
                sol = opt.minimize(fun = lfE.ll_ZINB, x0 = sol, method='Nelder-Mead',
                                                    args=(Wij,selection_variables,exogenous_variables,fixed_selection_params)).x
            except ZeroDivisionError:
                pass
            try:
                sol = opt.minimize(fun = lfE.ll_ZINB, jac = lfE.jac_ZINB,x0 = sol, method='BFGS',
                                                    args=(Wij,selection_variables,exogenous_variables,fixed_selection_params)).x
            except ZeroDivisionError:
                pass
            try:
                sol = opt.least_squares(fun = lfE.jac_ZINB, x0 = sol, 
                                                    args=(Wij,selection_variables,exogenous_variables,fixed_selection_params)).x
            except ZeroDivisionError:
                pass
            norm = np.linalg.norm(lfE.jac_ZINB(sol,Wij,selection_variables,exogenous_variables,fixed_selection_params),ord=np.inf)
            ll = - lfE.ll_ZINB(sol,Wij,selection_variables,exogenous_variables,fixed_selection_params)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm:',norm,'ll:',ll)
            
            
             
        return sol
    
    elif model == "CGeom":
        guess = solver("POIS",Wij,selection_variables,exogenous_variables,verbose=False)
        guess_b0 = np.ones(1)
        sol = np.concatenate((guess,guess_b0))
        result_linspace = []
        norm_linspace = []
        inf_bounds_ls = []
        sup_bounds_ls = []
        ll_linspace = []
        for i in range(len(sol)):
            inf_bounds_ls.append(-np.inf)
            sup_bounds_ls.append(np.inf)
        inf_bounds_ls[-1] = 1e-8
        
        bounds_ls = (inf_bounds_ls,sup_bounds_ls)
        step = 0
        norm = 10000
        while norm > tol and step < maxiter:
            
            try:
                sol = opt.minimize(fun=lfC.ll_CGeom,args=(Wij,exogenous_variables),x0 = sol, method='Nelder-Mead').x
            except:
                pass
            
            try:
                sol = opt.least_squares(fun=lfC.jac_CGeom,args=(Wij,exogenous_variables),x0 = sol,bounds=bounds_ls).x
            except:
                pass
            ll = lfC.ll_CGeom(sol,Wij,exogenous_variables)
            jac = lfC.jac_CGeom(sol,Wij,exogenous_variables)
            # print('iterazione',step,'jac',np.abs(jac))
            norm = np.linalg.norm(jac,ord=np.inf)
            
            result_linspace.append(sol)
            norm_linspace.append(norm)
            ll_linspace.append(ll)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm_w:',norm,'ll_w:',-ll)
                
            if norm < tol:
                if verbose:
                    print("Cond Geom: ",norm)
                    
                return sol 
                    
        minIndex = norm_linspace.index(min(norm_linspace))
        minIndex = ll_linspace.index(min(ll_linspace))
        
        true_result = result_linspace[minIndex]
        return true_result
    
    elif model == "CExp":
        sol = np.ones(exogenous_variables.shape[1]+1)
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
        
        step = 0
        norm = 10000
        while norm > tol and step < maxiter:
        
            sol = opt.minimize(fun=lfC.ll_CExp,method="Nelder-Mead",args=(Wij,exogenous_variables),x0 = sol).x
            
            sol = opt.least_squares(fun=lfC.jac_CExp,args=(Wij,exogenous_variables),x0 = sol,verbose=0).x
            ll = lfC.ll_CExp(sol,Wij,exogenous_variables)
            jac = lfC.jac_CExp(sol,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)    
            result_linspace.append(sol)
            norm_linspace.append(norm)
            ll_linspace.append(ll)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm_w:',norm,'ll_w:',-ll)
            
            
            if norm < tol:
                return sol
                    
        # minIndex = norm_linspace.index(min(norm_linspace))
        minIndex = ll_linspace.index(min(ll_linspace))
        
        true_result = result_linspace[minIndex]
            
        return true_result
            
    elif model == "CPareto":
        sol = solver("POIS",Wij,selection_variables,exogenous_variables,verbose=False)
        ll =  lfC.ll_CPareto(sol,Wij,exogenous_variables)
        # sol = np.zeros(exogenous_variables.shape[1])
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
        step = 0
        norm = 10000
        
        while norm > tol and step < maxiter:
            sol = opt.least_squares(fun=lfC.jac_CPareto,args=(Wij,exogenous_variables),
            x0 = sol).x
            ll = lfC.ll_CPareto(sol,Wij,exogenous_variables)
            jac = lfC.jac_CPareto(sol,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)    
            result_linspace.append(sol)
            norm_linspace.append(norm)
            ll_linspace.append(ll)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm_w:',norm,'ll_w:',-ll)
            
            if norm < tol:
                if verbose:
                    print("Cond Pareto: ", norm)
                return sol
                    
        # minIndex = norm_linspace.index(min(norm_linspace))
        minIndex = ll_linspace.index(min(ll_linspace))
        
        true_result = result_linspace[minIndex]
            
        return true_result
                
    elif model == "CLognormal":
        sol = np.ones(exogenous_variables.shape[1]+1)
        step = 0
        norm = 10000
        while norm > tol and step < maxiter:
            
            sol = opt.least_squares(fun=lfC.jac_CLognormal,x0=sol,args=(Wij,exogenous_variables),
                                    ).x
            jac = lfC.jac_CLognormal(sol,Wij,exogenous_variables)
            ll =  lfC.ll_CLognormal(sol,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm_w:',norm,'ll_w:',-ll)
            
        
        return sol

    elif model == "CGamma":
        sol = np.ones(exogenous_variables.shape[1]+2)
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
        step = 0
        norm = 10000
        while norm > tol and step < maxiter:
            try:    
                sol = opt.least_squares(fun=lfC.jac_CGamma,args=(Wij,exogenous_variables),x0 = sol,verbose=0).x
            except ZeroDivisionError:
                pass
            
            ll = lfC.ll_CGamma(sol,Wij,exogenous_variables)
            jac = lfC.jac_CGamma(sol,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)  
            # print('iterazione',it,'jac',jac)
            ll_linspace.append(ll)  
            result_linspace.append(sol)
            norm_linspace.append(norm)
            step += 1
            
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'norm_w:',norm,'ll_w:',-ll)
            
        
            if norm < tol:
                return sol
                    
        minIndex = norm_linspace.index(min(norm_linspace))
        # minIndex = ll_linspace.index(min(ll_linspace))
        true_result = result_linspace[minIndex]
            
        return true_result
                        
    elif model == "L-CGeom":
        n_selection_variables = selection_variables.shape[1] - len(fixed_selection_params)
        if len(use_guess)==0:
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,print_steps = print_steps,use_guess=np.array([]))
            
        else:  
            
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,print_steps = print_steps, use_guess=use_guess[:n_selection_variables])
            

        if len(use_guess) == 0:
            cond_Geom_params = solver("CGeom",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,verbose=verbose,print_steps = print_steps)
        else:
            cond_Geom_params = solver("CGeom",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,verbose=verbose,print_steps = print_steps,
                                      use_guess=use_guess[n_selection_variables:])
                
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
                
    elif model == "k-CGeom_undirected":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,print_steps = print_steps,maxiter=maxiter)
        else:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CGeom",Wij,selection_variables,exogenous_variables,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CGeom",Wij,selection_variables,exogenous_variables,maxiter=maxiter,verbose=verbose,print_steps = print_steps,use_guess=use_guess[n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
            
    elif model == "k-CGeom_directed":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params=np.array([1.00000,1.0000,0.0000]),tol=1e-5,
           use_guess= np.array([]),verbose=verbose,maxiter=maxiter)
        else:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params=np.array([1.00000,1.00000,0.0000]),tol=1e-5,
           use_guess= use_guess[:2*n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CGeom",Wij,selection_variables,exogenous_variables,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CGeom",Wij,selection_variables,exogenous_variables,maxiter=maxiter,verbose=verbose,print_steps = print_steps,
                                      use_guess=use_guess[2*n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
        
    elif model == "L-CExp":
        n_selection_variables = selection_variables.shape[1] - len(fixed_selection_params)
        if len(use_guess)==0:
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,
                                        print_steps = print_steps,use_guess=np.array([]))
            
        else:  
            
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,
                                        print_steps = print_steps,use_guess=use_guess[:n_selection_variables])
            

        if len(use_guess) == 0:
            cond_Geom_params = solver("CExp",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,
                                      print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CExp",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,
                                      print_steps = print_steps,verbose=verbose,
                                      use_guess=use_guess[n_selection_variables:])
                
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
                    
    elif model == "k-CExp_undirected":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,maxiter=maxiter)
        else:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CExp",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CExp",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,verbose=verbose,print_steps = print_steps,use_guess=use_guess[n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
                       
    elif model == "k-CExp_directed":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,print_steps = print_steps,maxiter=maxiter)
        else:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:2*n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CExp",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CExp",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,verbose=verbose,print_steps = print_steps,use_guess=use_guess[2*n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result

    elif model == "L-CPareto":
        n_selection_variables = selection_variables.shape[1] - len(fixed_selection_params)
        if len(use_guess)==0:
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,print_steps = print_steps,use_guess=np.array([]))
            
        else:  
            
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,print_steps = print_steps,use_guess=use_guess[:n_selection_variables])
            

        if len(use_guess) == 0:
            cond_Geom_params = solver("CPareto",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CPareto",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose,use_guess=use_guess[n_selection_variables:])
                
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
        
    elif model == "k-CPareto_undirected":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,print_steps = print_steps,maxiter=maxiter)
        else:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CPareto",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CPareto",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose,use_guess=use_guess[n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
    
    elif model == "k-CPareto_directed":
        
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,print_steps = print_steps,maxiter=maxiter)
        else:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:2*n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CPareto",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CPareto",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose,use_guess=use_guess[2*n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
        
        
        
    elif model == "L-CGamma":
        n_selection_variables = selection_variables.shape[1] - len(fixed_selection_params)
        if len(use_guess)==0:
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,print_steps = print_steps,verbose=verbose,use_guess=np.array([]))
            
        else:  
            
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,print_steps = print_steps,verbose=verbose,use_guess=use_guess[:n_selection_variables])
            

        if len(use_guess) == 0:
            cond_Geom_params = solver("CGamma",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CGamma",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose,use_guess=use_guess[n_selection_variables:])
                
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
    
    elif model == "k-CGamma_undirected":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,maxiter=maxiter)
        else:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:n_countries],verbose=verbose,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CGamma",Wij,selection_variables,exogenous_variables,fixed_selection_params,print_steps = print_steps,maxiter=maxiter,verbose=verbose)
        else:
            cond_Geom_params = solver("CGamma",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose,use_guess=use_guess[n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
    
    elif model == "k-CGamma_directed":
        
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,print_steps = print_steps,maxiter=maxiter)
        else:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:2*n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CGamma",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,verbose=verbose)
        else:
            cond_Geom_params = solver("CGamma",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,verbose=verbose,use_guess=use_guess[2*n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
        
    elif model == "L-CLognormal":
        n_selection_variables = selection_variables.shape[1] - len(fixed_selection_params)
        if len(use_guess)==0:
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,print_steps = print_steps,use_guess=np.array([]))
            
        else:  
            
            topological_params = solver("Logit",Wij,selection_variables,exogenous_variables,fixed_selection_params,verbose=verbose,print_steps = print_steps,use_guess=use_guess[:n_selection_variables])
            

        if len(use_guess) == 0:
            cond_Geom_params = solver("CLognormal",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CLognormal",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,verbose=verbose,print_steps = print_steps,use_guess=use_guess[n_selection_variables:])
                
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
    
    elif model == "k-CLognormal_undirected":
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,print_steps = print_steps,maxiter=maxiter)
        else:
            topological_params = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CLognormal",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CLognormal",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose,use_guess=use_guess[n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
    
    elif model == "k-CLognormal_directed":
        
        n_countries = int(np.sqrt(len(Wij)))
        if len(use_guess) == 0:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=verbose,print_steps = print_steps,maxiter=maxiter)
        else:
            topological_params = solver("DBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= use_guess[:2*n_countries],verbose=verbose,print_steps = print_steps,maxiter=maxiter)
            
        if len(use_guess) == 0:
            cond_Geom_params = solver("CLognormal",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose)
        else:
            cond_Geom_params = solver("CLognormal",Wij,selection_variables,exogenous_variables,fixed_selection_params,maxiter=maxiter,print_steps = print_steps,verbose=verbose,use_guess=use_guess[2*n_countries:])
            
        result = np.concatenate((topological_params,cond_Geom_params))
        return result
    
    elif model == "L-IGeom":
        guess_weighted = solver("POIS",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
           use_guess= np.array([]),verbose=False,maxiter=maxiter)
        guess_theta_0 = np.array([6.])      
        beta_linspace = np.linspace(0.99,1.,num=5)[:-1]
        result_linspace = []
        norm_linspace = []
        ll_linspace = []

        for lin in range(len(beta_linspace)):  
        
            guess_beta_0 = np.array([beta_linspace[lin]])
            sol = np.concatenate((guess_theta_0,guess_weighted,guess_beta_0))
            if len(use_guess)!=0:
                guess = use_guess
            norm = 10000
            step = 0
            while norm > tol and step < maxiter:    
                sol = opt.minimize(fun=lfI.ll_LIGeom,method="Nelder-Mead",args=(Wij,exogenous_variables),x0=sol).x
                sol = opt.least_squares(fun=lfI.jac_LIGeom,args=(Wij,exogenous_variables),x0 = sol).x
                jac = - lfI.jac_LIGeom(sol,Wij,exogenous_variables)
                ll =  lfI.ll_LIGeom(sol,Wij,exogenous_variables)
                norm = np.linalg.norm(jac,ord=np.inf)
                result_linspace.append(sol)
                norm_linspace.append(norm)
                ll_linspace.append(ll)
                step += 1
                if np.isnan(ll):
                    break
                if verbose:
                    if step % print_steps == 0:
                        print("round",lin,'iteration:',step,"norm",norm,'ll:',-ll)
                        
                if norm < tol:
                    return sol
                if len(norm_linspace) > 2:
                    if norm_linspace[-1] == norm_linspace[-2]:
                        break
                
                    
        minIndex = norm_linspace.index(min(norm_linspace))
        # minIndex = ll_linspace.index(min(ll_linspace))
        true_result = result_linspace[minIndex]
        true_norm = min(norm_linspace)
        true_jac = lfI.jac_LIGeom(true_result,Wij,exogenous_variables)
                    
        return true_result

    elif model == "L-IExp":
        guess_weighted = solver("POIS",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
        use_guess= np.array([]),verbose=False,maxiter=maxiter)
        guess_theta_0 = np.array([6.])      
        # beta_linspace = np.linspace(0.99,1.,num=5)[:-1]
        beta_linspace = np.array([0.99])
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
        
        norm = 10000
        step = 0
        
        for lin in range(len(beta_linspace)):  
            guess_beta_0 = np.array([beta_linspace[lin]])
            sol = np.concatenate((guess_theta_0,guess_weighted,guess_beta_0))
            
            norm = 10000
            step = 0
            while norm > tol and step < maxiter:
                
                sol = opt.minimize(fun=lfI.ll_LIExp,method="Nelder-Mead",args=(Wij,exogenous_variables),x0=sol).x
                sol = opt.least_squares(fun=lfI.jac_LIExp,args=(Wij,exogenous_variables),x0 = sol).x
                jac = lfI.jac_LIExp(sol,Wij,exogenous_variables)
                ll =  lfI.ll_LIExp(sol,Wij,exogenous_variables)
                norm = np.linalg.norm(jac,ord=np.inf)
                if np.isnan(ll):
                    break
                step += 1
                if verbose:
                    if step % print_steps == 0:
                        
                        print("round",lin,'iteration:',step,"norm",norm,'ll:',-ll)
                        
                if norm < tol:
                    if verbose:
                        print("LIA1: ",norm)
                    return sol
                result_linspace.append(sol)
                norm_linspace.append(norm)
                ll_linspace.append(ll)
                
                if len(norm_linspace) > 2:
                    if norm_linspace[-1] == norm_linspace[-2]:
                        break
                
                
        minIndex = ll_linspace.index(min(ll_linspace))
        true_result = result_linspace[minIndex]
                    
        return true_result

    elif model == "k-IGeom_undirected":
        n_countries = int(np.sqrt(len(Wij)))
        n_select_variables = n_countries
        Aij = binarize(Wij)
        
        if len(use_guess) == 0:
            guess_theta_0 = solver("UBCM",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
                    use_guess= np.array([]),verbose=False,maxiter=maxiter)
            guess_other = solver("L-IGeom",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
                    use_guess= np.array([]),verbose=False,maxiter=maxiter)
            guess = np.concatenate((guess_theta_0,guess_other[1:]))
            
        if len(use_guess) != 0:
            guess = use_guess
        
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
        
        norm = 100000
        step = 0
        while step < maxiter and norm > tol:
            
            try:
                guess = opt.minimize(fun=lfI.ll_kIGeom_undirected,jac=lfI.jac_kIGeom_undirected,method="BFGS",
                                        args=(Wij,exogenous_variables),x0=guess).x
            except ZeroDivisionError:
                pass
            
            
            try:
                guess_beta_0 = np.array([guess[-1]])
                guess_theta_0 = guess[:n_countries]
                guess_weighted = guess[n_countries:-1]
                theta_0 = opt.least_squares(fun=lfI.jac_kIGeom_topological_undirected,jac=lfI.hess_kIGeom_topological_undirected,
                                                x0 = guess_theta_0, verbose=0,
                                            args=(Wij,exogenous_variables, guess_weighted,guess_beta_0))
                
                guess = np.concatenate((theta_0.x,guess_weighted,guess_beta_0))
            except ZeroDivisionError:
                pass
            
            
            
            try:
                guess = opt.least_squares(fun=lfI.jac_kIGeom_undirected,args=(Wij,exogenous_variables),
                                                    x0 = guess,verbose=0,max_nfev=100).x
            except ZeroDivisionError:
                pass
            guess_theta_0 = guess[:n_countries]
            guess_weighted = guess[n_countries:-1]
            guess_beta_0 = np.array([guess[-1]])
            
            jac = lfI.jac_kIGeom_undirected(guess,Wij,exogenous_variables)
            ll =  lfI.ll_kIGeom_undirected(guess,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)
            norm_top = np.linalg.norm(jac[:n_countries],ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'ll:',-ll,'norm_top:',norm_top,'norm:',norm)
                    
            if np.isnan(ll):
                break
                    
            if norm < tol:
                return guess
            result_linspace.append(guess)
            norm_linspace.append(norm)
            ll_linspace.append(-ll)
            if len(norm_linspace) > 2:
                if norm_linspace[-1] == norm_linspace[-2]:
                    break
                    
                        
        minIndex = ll_linspace.index(min(ll_linspace))
        true_result = result_linspace[minIndex]
                    
        return true_result

    elif model == "k-IGeom_directed":
        n_countries = int(np.sqrt(len(Wij)))
        n_select_variables = n_countries
        Aij = binarize(Wij)
        
        if len(use_guess) == 0:
            guess_theta_0 = solver("DBCM",Wij,selection_variables,exogenous_variables,tol=1e-5,
                    use_guess= np.array([]),verbose=False,maxiter=maxiter)
            guess_other = solver("L-IGeom",Wij,selection_variables,exogenous_variables,tol=1e-5,
                    use_guess= np.array([]),verbose=False,maxiter=maxiter)
            guess = np.concatenate((guess_theta_0,guess_other[1:]))
            
            
            
            
        if len(use_guess) != 0:
            guess = use_guess
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
            
        norm = 100000
        step = 0
        while norm > tol and step < maxiter:
        
            
            try:
                guess = opt.minimize(fun=lfI.ll_kIGeom_directed,jac=lfI.jac_kIGeom_directed,method="BFGS",
                                        args=(Wij,exogenous_variables),x0=guess).x
            except ZeroDivisionError:
                pass
            
            
            try:
                guess_beta_0 = np.array([guess[-1]])
                guess_theta_0 = guess[:2*n_countries]
                guess_weighted = guess[2*n_countries:-1]
                theta_0 = opt.least_squares(fun=lfI.jac_kIGeom_topological_directed,
                                                x0 = guess_theta_0, verbose=0,
                                            args=(Wij,exogenous_variables, guess_weighted,guess_beta_0))
                
                guess = np.concatenate((theta_0.x,guess_weighted,guess_beta_0))
            except ZeroDivisionError:
                pass
            
            
            
            try:
                guess = opt.least_squares(fun=lfI.jac_kIGeom_directed,args=(Wij,exogenous_variables),
                                                    x0 = guess,verbose=0,max_nfev=100).x
            except ZeroDivisionError:
                pass
            guess_theta_0 = guess[:2*n_countries]
            guess_weighted = guess[2*n_countries:-1]
            guess_beta_0 = np.array([guess[-1]])
            
            jac = lfI.jac_kIGeom_directed(guess,Wij,exogenous_variables)
            ll =  lfI.ll_kIGeom_directed(guess,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)
            norm_top = np.linalg.norm(jac[:n_countries],ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'ll:',-ll,'norm_top:',norm_top,'norm:',norm)
                    
            if np.isnan(ll):
                break
                    
            if norm < tol:
                return guess
            result_linspace.append(guess)
            norm_linspace.append(norm)
            ll_linspace.append(-ll)
            if len(norm_linspace) > 2:
                if norm_linspace[-1] == norm_linspace[-2]:
                    break
                    
                        
        minIndex = norm_linspace.index(min(norm_linspace))
        minIndex = ll_linspace.index(min(ll_linspace))
        
        true_result = result_linspace[minIndex]
                    
        return true_result
    
    
    elif model == "k-IExp_undirected":
        n_countries = int(np.sqrt(len(Wij)))
        n_select_variables = n_countries
        Aij = binarize(Wij)
        if len(use_guess) == 0:
            Exp_TS = solver("k-CExp_undirected",Wij,selection_variables,exogenous_variables,fixed_selection_params,tol=1e-5,
                use_guess= np.array([]),verbose=False,maxiter=maxiter)
            guess_weighted = Exp_TS[n_countries:-1]
            guess_theta_0 = Exp_TS[:n_countries]
            guess = Exp_TS
        if len(use_guess) != 0:
            guess = use_guess
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
        norm = 10000
        step = 0
        while norm > tol and step < maxiter:
            
            try:
                guess_beta_0 = np.array([guess[-1]])
                guess_theta_0 = guess[:n_countries]
                guess_weighted = guess[n_countries:-1]
                theta_0 = opt.least_squares(fun=lfI.jac_kIExp_topological_undirected,
                                                x0 = guess_theta_0, verbose=0,
                                            args=(Wij,exogenous_variables, guess_weighted,guess_beta_0))
                
                guess = np.concatenate((theta_0.x,guess_weighted,guess_beta_0))
            except ZeroDivisionError:
                pass    
            try:
                guess = opt.minimize(fun=lfI.ll_kIExp_undirected,jac=lfI.jac_kIExp_undirected,method='BFGS',
                                            args=(Wij,exogenous_variables),x0=guess).x
            except ZeroDivisionError:
                pass
            
            try:
                guess = opt.least_squares(fun=lfI.jac_kIExp_undirected,args=(Wij,exogenous_variables),
                                                    x0 = guess,verbose=0,max_nfev=100).x
            except ZeroDivisionError:
                pass    
            
            # guess = result_minimize.x
            guess_theta_0 = guess[:n_countries]
            guess_weighted = guess[n_countries:-1]
            guess_beta_0 = np.array([guess[-1]])
            
            jac = lfI.jac_kIExp_undirected(guess,Wij,exogenous_variables)
            jac_top = lfI.jac_kIExp_topological_undirected(guess_theta_0,Wij,exogenous_variables,guess_weighted,guess_beta_0)
            
            ll =  lfI.ll_kIExp_undirected(guess,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)
            norm_top = np.linalg.norm(jac_top,ord=np.inf)
            
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'ll:',-ll,'norm_top:',norm_top,'norm:',norm)
                    
            if np.isnan(ll):
                break
            if norm < tol:
                return guess
            result_linspace.append(guess)
            norm_linspace.append(norm)
            ll_linspace.append(ll)
            if len(norm_linspace) > 2:
                if norm_linspace[-1] == norm_linspace[-2]:
                    break
                    
                        
        minIndex = ll_linspace.index(min(ll_linspace))
        true_result = result_linspace[minIndex]
        true_norm = min(norm_linspace)
                    
        return true_result
    
    elif model == "k-IExp_directed":
        n_countries = int(np.sqrt(len(Wij)))
        n_select_variables = n_countries
        Aij = binarize(Wij)
        
        if len(use_guess) == 0:
            guess_theta_0 = solver("DBCM",Wij,selection_variables,exogenous_variables,tol=1e-5,
                    use_guess= np.array([]),verbose=False,maxiter=maxiter)
            guess_other = solver("L-IExp",Wij,selection_variables,exogenous_variables,tol=1e-5,
                    use_guess= np.array([]),verbose=False,maxiter=maxiter)
            guess = np.concatenate((guess_theta_0,guess_other[1:]))
            
            
            
            
        if len(use_guess) != 0:
            guess = use_guess
        result_linspace = []
        norm_linspace = []
        ll_linspace = []
    
        norm = 10000
        step = 0
        while norm > tol and step < maxiter:
                
            try:
                guess = opt.minimize(fun=lfI.ll_kIExp_directed,jac=lfI.jac_kIExp_directed,method="BFGS",
                                        args=(Wij,exogenous_variables),x0=guess).x
            except ZeroDivisionError:
                pass
            
            
            try:
                guess_beta_0 = np.array([guess[-1]])
                guess_theta_0 = guess[:2*n_countries]
                guess_weighted = guess[2*n_countries:-1]
                theta_0 = opt.least_squares(fun=lfI.jac_kIExp_topological_directed,
                                                x0 = guess_theta_0, verbose=0,
                                            args=(Wij,exogenous_variables, guess_weighted,guess_beta_0))
                
                guess = np.concatenate((theta_0.x,guess_weighted,guess_beta_0))
            except ZeroDivisionError:
                pass
            
            
            
            try:
                guess = opt.least_squares(fun=lfI.jac_kIExp_directed,args=(Wij,exogenous_variables),
                                                    x0 = guess,verbose=0,max_nfev=100).x
            except ZeroDivisionError:
                pass
            guess_theta_0 = guess[:2*n_countries]
            guess_weighted = guess[2*n_countries:-1]
            guess_beta_0 = np.array([guess[-1]])
            
            jac = lfI.jac_kIExp_directed(guess,Wij,exogenous_variables)
            ll =  lfI.ll_kIExp_directed(guess,Wij,exogenous_variables)
            norm = np.linalg.norm(jac,ord=np.inf)
            norm_top = np.linalg.norm(jac[:n_countries],ord=np.inf)
            step += 1
            if verbose:
                if step % print_steps == 0:
                    print('iteration:',step,'ll:',-ll,'norm_top:',norm_top,'norm:',norm)
                    
            if np.isnan(ll):
                break
                    
            if norm < tol:
                return guess
            result_linspace.append(guess)
            norm_linspace.append(norm)
            ll_linspace.append(-ll)
            
            if len(norm_linspace) > 2:
                if norm_linspace[-1] == norm_linspace[-2]:
                    break
                    
                        
        minIndex = norm_linspace.index(min(norm_linspace))
        minIndex = ll_linspace.index(min(ll_linspace))
        
        true_result = result_linspace[minIndex]
                    
        return true_result
    

    else:
        raise TypeError('model wrongly defined. See the available models using .implemented_models')
            
            



import time
import numpy as np
from numba import jit, float64
import math


@jit(nopython=True,fastmath=True)
def binarize(input):
    """Computes binary adjacency matrix from weighted matrix.
    """
    
    dim = len(input)
    byn = np.zeros(dim)
    for i in range(dim):
        if input[i]> 0:
            byn[i] = 1.
    return byn 



@jit(nopython=True,fastmath=True)
def ll_LIGeom(params,Wij,exogenous_variables):
    """Computes the opposite of the log-likelihood for the L-constrained Integrated Geometric model"""
    params_exog = params[1:-1]
    
    x_0 = np.exp(-params[0])
    z_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
    
                yi = Wij[ij]
                ai = Aij[ij]

                ll += yi*np.log(z_0) + ai*np.log(x_0)
                ll += yi * np.log(mu[ij]) - yi * np.log(1+mu[ij]) + np.log(1+ mu[ij] - z_0*mu[ij]) - np.log(1 + mu[ij] - z_0*mu[ij]+ x_0*z_0*mu[ij])

                
    return - ll 


@jit(nopython=True,fastmath=True)
def ll_LIGeom_binary(params,Wij,exogenous_variables):
    """Computes the opposite of the binary log-likelihood for the L-constrained Integrated Geometric model"""
    
    params_exog = params[1:-1]
    
    x_0 = np.exp(-params[0])
    z_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    zij = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
    
                yi = Wij[ij]
                ai = Aij[ij]
                pij = x_0 * z_0 * zij[ij]/(1+zij[ij]-z_0*zij[ij]+x_0*z_0*zij[ij])
                ll += ai*np.log(pij) + (1.-ai)*np.log(1.-pij)
                
                
    return - ll 

    


@jit(nopython=True,fastmath=True)
def jac_LIGeom(params,Wij,exogenous_variables):
    """Computes the opposite of the jacobian for the L-constrained Integrated Geometric model"""
    
    params_exog = params[1:-1]
    theta_0 = params[0]
    x_0 = np.exp(-theta_0)
    z_0 = params[-1]
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)

    n_obs = len(Aij)
    jac = np.zeros(len(params))

    mu_matrix = np.zeros(len(Aij))
    
    adder_0 = 0
    adder_last = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                pi = x_0 * z_0 * zi/(1+zi - z_0*zi + x_0*z_0*zi)
                den_2 = (1+zi-z_0*zi)*(1+zi-z_0*zi+x_0*z_0*zi)
                yi_hat = x_0*z_0*zi*(1+zi)/den_2

                adder_0 += pi
                adder_last += yi_hat
                mu_matrix[ij] = (yi-yi_hat)/(1+zi)
                
    jac[0] = adder_0 - Aij.sum()
    jac[1:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = (adder_last - Wij.sum())
    
    return - jac




@jit(nopython=True,fastmath=True)
def hess_LIGeom(params,Wij,exogenous_variables):
    """Computes the opposite of the hessian for the L-constrained Integrated Geometric model"""
    
    n_exog_variables = exogenous_variables.shape[1]
    n_countries = int(np.sqrt(len(Wij)))
    params_exog = params[1:-1]
    theta_0 = params[0]
    x_0 = np.exp(-theta_0)
    z_0 = params[-1]
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    Aij = binarize(Wij)
    n_obs = len(Aij)
    hess = np.zeros((len(params),len(params)))
    
    W_matrix = np.zeros((len(Wij),len(Wij)))
    aux_theta_zij = np.zeros(len(Wij))
    aux_zij_z_0 = np.zeros(len(Wij))
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                aij = Aij[ij]
                wij = Wij[ij]
                zij = mu[ij]
                pij = x_0*z_0*zij/(1.+zij-z_0*zij+x_0*z_0*zij)
                den_1 = (1.+zij-z_0*zij+x_0*z_0*zij)
                den_2 = (1.+zij-z_0*zij)*(1.+zij-z_0*zij+x_0*z_0*zij)
                yi_hat = x_0*z_0*zij*(1+zij)/den_2

                d_den_2_z_0 = -zij*(1.+zij-z_0+x_0*z_0*zij) + (1.+zij-z_0*zij)*(-zij + x_0 * zij)
                d_den_2_zij = ((1.-z_0)*(1.+zij-z_0+x_0*z_0*zij) + (1.+zij-z_0*zij) * (1.- z_0 + x_0*z_0))*zij


                hess[0,0] += - pij*(1.-pij)
                aux_theta_zij[ij] = x_0*z_0*zij/(den_1)**2 
                hess[0,-1] += - x_0*z_0*zij/(den_1)**2
                hess[-1,0] += - x_0*z_0*zij/(den_1)**2
                
                num_W_matrix = x_0*z_0*den_2 - x_0*z_0*zij*d_den_2_zij
                W_matrix[ij,ij] = (- wij/(1+zij)**2 -  num_W_matrix/den_2**2)*zij

                num_zij_z_0 = x_0*zij*den_2 - x_0*z_0*zij*d_den_2_z_0
                aux_zij_z_0[ij] = - z_0*num_zij_z_0/den_2**2

                num_z_0_z_0 = x_0*zij*(1+zij)*den_2 - x_0*z_0*zij*(1+zij)*d_den_2_z_0
                hess[-1,-1] += - z_0*num_z_0_z_0/den_2**2

    hess[0,1:-1] = exogenous_variables.T @ aux_theta_zij
    hess[1:-1,0] = exogenous_variables.T @ aux_theta_zij
    hess[1:-1,1:-1] = exogenous_variables.T @ W_matrix @ exogenous_variables
    hess[-1,1:-1] = exogenous_variables.T @ aux_zij_z_0
    hess[1:-1,-1] = exogenous_variables.T @ aux_zij_z_0
    
    return - hess
    






@jit(float64(float64[:],float64[:],float64[:,:]),nopython=True,fastmath=False)
def ll_kIGeom_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the log-likelihood for the k-constrained Integrated Geometric model for Undirected networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    params_exog = params[n_countries:-1]
    z_0 = params[-1]
    theta_i = params[:n_countries]
    x_i = np.exp(-theta_i)

    exogenous_variables = np.ascontiguousarray(exogenous_variables)
    params_exog = np.ascontiguousarray(params_exog)
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    
    ll = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                mui = mu[ij]
                xij = x_i[i]*x_i[j]
                ll += ai*np.log(xij)  + yi*np.log(z_0) 
                ll += yi * np.log(mui) - yi * np.log(1.+mui) + np.log(1.+ mui - z_0*mui) - np.log(1. + mui - z_0*mui+ xij*z_0*mui)

    return - ll 

@jit(nopython=True,fastmath=True)
def ll_kIGeom_binary_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the binary log-likelihood for the k-constrained Integrated Geometric model for Undirected networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    params_exog = params[n_countries:-1]
    z_0 = params[-1]
    theta_i = params[:n_countries]
    x_i = np.exp(-theta_i)

    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    
    ll = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                mui = mu[ij]
                xij = x_i[i]*x_i[j]
                ll += ai*np.log(xij*z_0*mui) + (1.-ai)*np.log(1+ mui - z_0*mui) - np.log(1 + mui - z_0*mui+ xij*z_0*mui)

    return - ll 
    


@jit(float64[:](float64[:],float64[:],float64[:,:]),nopython=True,fastmath=True)
def jac_kIGeom_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the jacobian for the k-constrained Integrated Geometric model for Undirected networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    params_exog = params[n_countries:-1]
    
    
    theta_i = params[:n_countries]
    x_i = np.exp(-theta_i)
    z_0 = params[-1]
    exogenous_variables = np.ascontiguousarray(exogenous_variables)
    params_exog = np.ascontiguousarray(params_exog)
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    jac = np.zeros(len(params))
    
    mu_matrix = np.zeros(len(Aij))
    adder_last = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                mui = mu[ij]
                xij = x_i[i]*x_i[j]
                pi = xij*z_0*mui/(1.+mui-z_0*mui+xij*z_0*mui)
                jac[i] += - ai + pi

                
                denominator_mu = (1.+mui-z_0*mui+xij*z_0*mui) * (1.+mui-z_0*mui)
                mu_matrix[ij] = yi/(1.+mui) - xij*z_0*mui/denominator_mu

                adder_last += - yi  + xij*z_0*mui*(1.+mui)/denominator_mu
    
    jac[n_countries:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder_last
    return - jac

@jit(float64[:](float64[:],float64[:],float64[:,:],float64[:],float64[:]),nopython=True,fastmath=True)
def jac_kIGeom_topological_undirected(theta_i,Wij,exogenous_variables,params_exog,z_0):
    """Computes the opposite of the topological jacobian for the k-constrained Integrated Geometric model for Undirected networks.
    Used in the solver."""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    
    z_0 = z_0[0]
    x_i = np.exp(-theta_i)
    exogenous_variables = np.ascontiguousarray(exogenous_variables)
    params_exog = np.ascontiguousarray(params_exog)
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    jac = np.zeros(len(x_i))
    
    adder_last = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                mui = mu[ij]
                xij = x_i[i]*x_i[j]
                jac[i] +=  - ai + xij*z_0*mui/(1+mui-z_0*mui+xij*z_0*mui)
    
    return - jac




@jit(float64[:,:](float64[:],float64[:],float64[:,:]),nopython=True,fastmath=True)
def hess_kIGeom_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the hessian for the k-constrained Integrated Geometric model for Undirected networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    theta = params[:n_countries]
    params_exog = params[n_countries:-1]
    z_0 = params[-1]
    
    Aij = binarize(Wij)
    hess = np.zeros((len(params),len(params)))
    exogenous_variables = np.ascontiguousarray(exogenous_variables)
    params_exog = np.ascontiguousarray(params_exog)
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    W_matrix = np.zeros((len(Wij),len(Wij)))
    aux_theta_zij = np.zeros(len(Wij))
    aux_zij_z_0 = np.zeros(len(Wij))
    for i in range(n_countries):
        for j in range(n_countries):
            
            if j!=i:
                ij = i*n_countries+j
                xij = np.exp(-theta[i]-theta[j])
                
                aij = Aij[ij]
                wij = Wij[ij]
                zij = mu[ij]
                
                
                den_2 = (1+zij-z_0*zij)*(1+zij-z_0*zij+xij*z_0*zij)
                d_den_2_z_0 = - 2*zij + xij*zij - 2*zij**2 + xij*zij**2 +2*z_0*zij**2 - 2*xij*z_0*zij**2
                d_den_2_zij = 2 -2*z_0 + xij*z_0 + 2*zij - 4*z_0*zij + 2*xij*z_0*zij + 2*z_0**2 * zij - 2*xij*z_0**2 * zij
                
                diff_d_den_z_0 = xij*z_0*zij**2 + 1 + zij**2 -z_0**2 * zij**2 + 2*zij
                diff_d_den_zij = 1 - zij**2 - (z_0 * zij)**2 + 2*z_0*zij**2 - xij*z_0*zij**2 + xij*(z_0 * zij)**2
                
                
                pij = xij * zij * z_0/(1+zij -z_0*zij + xij*z_0*zij)
                wij_hat = xij*zij*z_0*(1+zij) /den_2
                
                d_a_theta_i_theta_i = - pij*(1-pij)
                d_a_theta_i_theta_j = - pij*(1-pij)
                d_a_theta_i_zij = -xij*z_0/(1+zij -z_0*zij + xij*z_0*zij)**2
                d_a_theta_i_b0 = + xij*zij*(1+zij)*z_0/(1+zij -z_0*zij + xij*z_0*zij)**2
                hess[i,i] += d_a_theta_i_theta_i
                hess[i,j] = d_a_theta_i_theta_j
                hess[j,i] = d_a_theta_i_theta_j
                hess[i,-1] += d_a_theta_i_b0 
                hess[-1,i] += d_a_theta_i_b0
                
                
                aux_theta_zij[ij] = d_a_theta_i_zij*zij 
                d_a_phi_zij = - wij/(1+zij)**2 - xij*z_0*diff_d_den_zij/den_2**2
                d_a_phi_b0 = - xij*zij*z_0*diff_d_den_z_0/den_2**2

                d_a_b0_b0 = - xij*zij*z_0*(1+zij)*diff_d_den_z_0/den_2**2
                
                W_matrix[ij,ij] = d_a_phi_zij*zij
                aux_zij_z_0[ij] = d_a_phi_b0
                hess[-1,-1] += d_a_b0_b0

    for i in range(n_countries):
        hess[i,n_countries:-1] = exogenous_variables.T @ aux_theta_zij
        hess[n_countries:-1,i] = exogenous_variables.T @ aux_theta_zij       
    hess[n_countries:-1,n_countries:-1] = exogenous_variables.T @ W_matrix @ exogenous_variables
    hess[-1,n_countries:-1] = exogenous_variables.T @ aux_zij_z_0
    hess[n_countries:-1,-1] = exogenous_variables.T @ aux_zij_z_0
                

                                 
    return - hess




@jit(float64[:,:](float64[:],float64[:],float64[:,:],float64[:],float64[:]),nopython=True,fastmath=True)
def hess_kIGeom_topological_undirected(theta,Wij,exogenous_variables,params_exog,z_0):
    """Computes the opposite of the topological hessian for the k-constrained Integrated Geometric model for Undirected networks.
    Used in the solver"""
    
    n_countries = int(np.sqrt(len(Wij)))
    z_0 = z_0[0]
    Aij = binarize(Wij)
    hess = np.zeros((len(theta),len(theta)))
    exogenous_variables = np.ascontiguousarray(exogenous_variables)
    params_exog = np.ascontiguousarray(params_exog)
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries+j
                xij = np.exp(-theta[i]-theta[j])
                
                aij = Aij[ij]
                wij = Wij[ij]
                zij = mu[ij]
                
                pij = xij * zij * z_0/(1+zij -z_0*zij + xij*z_0*zij)
                
                d_a_theta_i_theta_i = - pij*(1-pij)
                d_a_theta_i_theta_j = - pij*(1-pij)
                hess[i,i] += d_a_theta_i_theta_i
                hess[i,j] = d_a_theta_i_theta_j
                hess[j,i] = d_a_theta_i_theta_j
                                 
    return - hess


@jit(nopython=True,fastmath=True)
def ll_kIGeom_directed(params,Wij,exogenous_variables):
    """Computes the opposite of the log-likelihood for the k-constrained Integrated Geometric model for Directed networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    params_exog = params[2*n_countries:-1]
    z_0 = params[-1]
    theta_out = params[:n_countries]
    theta_in = params[n_countries:2*n_countries]
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)

    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    
    ll = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                mui = mu[ij]
                xij = x_out[i]*x_in[j]
                ll += ai*np.log(xij)  + yi*np.log(z_0) 
                ll += yi * np.log(mui) - yi * np.log(1+mui) + np.log(1+ mui - z_0*mui) - np.log(1 + mui - z_0*mui+ xij*z_0*mui)

    return - ll 

@jit(nopython=True,fastmath=True)
def ll_kIGeom_binary_directed(params,Wij,exogenous_variables):
    """Computes the opposite of the binary log-likelihood for the k-constrained Integrated Geometric model for Directed networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    params_exog = params[2*n_countries:-1]
    z_0 = params[-1]
    theta_out = params[:n_countries]
    theta_in = params[n_countries:2*n_countries]
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)

    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    
    ll = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                mui = mu[ij]
                xij = x_out[i]*x_in[j]
                ll += ai*np.log(xij*z_0*mui) + (1.-ai)*np.log(1+ mui - z_0*mui) - np.log(1 + mui - z_0*mui+ xij*z_0*mui)

    return - ll 
    


@jit(nopython=True,fastmath=True)
def jac_kIGeom_directed(params,Wij,exogenous_variables):
    """Computes the opposite of the jacobian for the k-constrained Integrated Geometric model for Directed networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    params_exog = params[2*n_countries:-1]
    z_0 = params[-1]
    theta_out = params[:n_countries]
    theta_in = params[n_countries:2*n_countries]
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)

    

    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    jac = np.zeros(len(params))
    
    mu_matrix = np.zeros(len(Aij))
    adder_last = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                ji = j*n_countries + i
                
                yi = Wij[ij]
                aij = Aij[ij]
                aji = Aij[ji]
                muij = mu[ij]
                muji = mu[ji]
                xij = x_out[i]*x_in[j]
                pij = xij*z_0*muij/(1+muij-z_0*muij+xij*z_0*muij)
                xji = x_out[j]*x_in[i]
                pji = xji*z_0*muji/(1+muji-z_0*muji+xji*z_0*muji)
                
                jac[i] += - aij + pij
                jac[i+n_countries] += - aji + pji

                
                denominator_mu = (1+muij-z_0*muij+xij*z_0*muij) * (1+muij-z_0*muij)
                mu_matrix[ij] = yi/(1+muij) - xij*z_0*muij/denominator_mu

                adder_last += - yi  + xij*z_0*muij*(1+muij)/denominator_mu
    
    jac[2*n_countries:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder_last
    return - jac

@jit(nopython=True,fastmath=True)
def jac_kIGeom_topological_directed(theta,Wij,exogenous_variables,params_exog,z_0):
    """Computes the opposite of the topological jacobian for the k-constrained Integrated Geometric model for Directed networks"""
    
    n_countries = int(np.sqrt(len(Wij)))
    n_obs = len(Wij)
    Aij = binarize(Wij)
    
    
    z_0 = z_0[0]
    theta_out = theta[:n_countries]
    theta_in = theta[n_countries:2*n_countries]
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)

    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    jac = np.zeros(len(theta))
    
    adder_last = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                mui = mu[ij]
                xij = x_out[i]*x_in[j]
                jac[i] +=  - ai + xij*z_0*mui/(1+mui-z_0*mui+xij*z_0*mui)
    
    return - jac






@jit(nopython=True,fastmath=True)
def ll_LIExp(params,Wij,exogenous_variables):
    """Computes the opposite of the log-likelihood for the L-constrained Integrated Exponential model"""
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[1:-1]
    
    x_0 = np.exp(-params[0])
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
    
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                bij = beta_0 + 1./zi
                ll += ai*np.log(x_0) + np.log(bij) - bij*yi - np.log(x_0+bij)
                
    return - ll 

@jit(nopython=True,fastmath=True)
def ll_LIExp_binary(params,Wij,exogenous_variables):
    """Computes the opposite of the binary log-likelihood for the L-constrained Integrated Exponential model"""
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[1:-1]
    
    x_0 = np.exp(-params[0])
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
    
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                pij = x_0 * mu[ij]/(1+beta_0*mu[ij]+x_0*mu[ij])

                ll += ai*np.log(pij) + (1.-ai)*np.log(1.-pij)
                
    return - ll 

@jit(nopython=True,fastmath=True)
def jac_LIExp(params,Wij,exogenous_variables):
    """Computes the opposite of the jacobian for the L-constrained Integrated Exponential model"""
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[1:-1]
    theta_0 = params[0]
    x_0 = np.exp(-theta_0)
    beta_0 = params[-1]
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)

    n_obs = len(Aij)
    jac = np.zeros(len(params))

    mu_matrix = np.zeros(len(Aij))
    
    adder_0 = 0
    adder_last = 0
    n_countries = int(np.sqrt(len(Aij)))
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                pi = x_0 * zi/(1+ beta_0*zi + x_0*zi)
                yi_hat = pi*zi/(1+beta_0*zi)

                adder_0 += pi - ai
                adder_last += yi_hat - yi
                mu_matrix[ij] = (yi-yi_hat)/(zi)
                
    jac[0] = adder_0 
    jac[1:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder_last
    
    return - jac






@jit(nopython=True,fastmath=True)
def hess_LIExp(params,Wij,exogenous_variables):
    """Computes the opposite of the hessian for the L-constrained Integrated Exponential model"""
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[1:-1]
    theta_0 = params[0]
    x = np.exp(-theta_0)
    b0 = params[-1]
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)

    n_obs = len(Aij)
    hess = np.zeros((len(params),len(params)))

    mu_matrix = np.zeros(len(Aij))
    
    adder_0 = 0
    adder_last = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    d_x_zij = np.zeros(n_obs)
    d_zij_zij = np.zeros((n_obs,n_obs))
    d_zij_b0 = np.zeros(n_obs)
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries+j
                aij = Aij[ij]
                wij = Wij[ij]
                zij = mu[ij]
                pij = x *zij/(1+zij*(b0+zij))
                den_1 = 1. + b0*zij + x*zij
                den_2 = den_1 * (1+b0*zij)
                
                hess[0,0] += - pij*(1-pij)
                hess[0,-1] += - x*zij**2/den_1**2
                hess[-1,0] += - x*zij**2/den_1**2
                d_x_zij[ij] = x*zij/den_1**2

                diffin = (b0+x)*(1.+b0*zij) + b0*(1.+b0*zij+x*zij)
                d_zij_zij[ij,ij] = -wij/zij - x*zij**2/den_2**2 + x*zij**2 * diffin/den_2**2
                d_zij_b0[ij] = x*zij**2 *(2.*zij*(1+b0*zij)+x*zij**2)/den_2**2

                hess[-1,-1] += - x*zij**2 * (2.*zij*(1.+b0*zij) + x*zij**2)/den_2**2
                
    hess[0,1:-1] = exogenous_variables.T @ d_x_zij
    hess[1:-1,0] = exogenous_variables.T @ d_x_zij
    hess[1:-1,1:-1] = exogenous_variables.T @ d_zij_zij @ exogenous_variables
    hess[1:-1,-1] = exogenous_variables.T @ d_zij_b0
    hess[-1,1:-1] = exogenous_variables.T @ d_zij_b0

    return - hess


@jit(nopython=True,fastmath=True)
def ll_kIExp_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the log-likelihood for the k-constrained Integrated Exponential model for Undirected networks"""
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[n_countries:-1]
    theta = params[:n_countries]
    x_i = np.exp(-theta)
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                xij = x_i[i]*x_i[j]
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]

                bij = beta_0 + 1./zi
                ll += ai*np.log(xij) + np.log(bij) - bij*yi - np.log(xij+bij)
                
                
    return - ll 



@jit(nopython=True,fastmath=True)
def ll_kIExp_binary_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the binary log-likelihood for the k-constrained Integrated Exponential model for Undirected networks"""
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[n_countries:-1]
    theta = params[:n_countries]
    x_i = np.exp(-theta)
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                xij = x_i[i]*x_i[j]
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                ll += ai*np.log(xij*zi) + (1.-ai)*np.log(1.+beta_0*zi) - np.log(1.+beta_0*zi + xij*zi) 
                

    return - ll 


@jit(nopython=True,fastmath=True)
def jac_kIExp_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the jacobian for the k-constrained Integrated Exponential model for Undirected networks"""
    
    Aij = binarize(Wij)
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[n_countries:-1]
    theta = params[:n_countries]
    x_i = np.exp(-theta)
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    jac = np.zeros(len(params))

    mu_matrix = np.zeros(len(Aij))
    
    adder_last = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                xij = x_i[i]*x_i[j]
                pi = xij * zi/(1+ beta_0*zi + xij*zi)
                yi_hat = pi*zi/(1+beta_0*zi)

                jac[i] += - ai + pi
                adder_last += yi_hat - yi
                mu_matrix[ij] = (yi-yi_hat)/(zi)
                
    jac[n_countries:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder_last 
    
    return - jac

@jit(nopython=True,fastmath=True)
def jac_kIExp_topological_undirected(theta,Wij,exogenous_variables,params_exog,beta_0):
    """Computes the opposite of the topological jacobian for the k-constrained Integrated Exponential model for Undirected networks"""
    
    Aij = binarize(Wij)
    n_countries = int(np.sqrt(len(Aij)))
    
    x_i = np.exp(-theta)
    beta_0 = beta_0[0]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    jac = np.zeros(len(theta))

    mu_matrix = np.zeros(len(Aij))
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                xij = x_i[i]*x_i[j]
                pi = xij * zi/(1+ beta_0*zi + xij*zi)
                yi_hat = pi*zi/(1+beta_0*zi)

                jac[i] += - ai + pi
                
    
    return - jac



@jit(nopython=True,fastmath=True)
def hess_kIExp_undirected(params,Wij,exogenous_variables):
    """Computes the opposite of the hessian for the k-constrained Integrated Exponential model for Undirected networks"""
    
    
    n_countries = int(np.sqrt(len(Wij)))
    params_exog = params[n_countries:-1]
    theta_0 = params[:n_countries]
    x = np.exp(-theta_0)
    b0 = params[-1]
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    Aij = binarize(Wij)

    n_obs = len(Aij)
    hess = np.zeros((len(params),len(params)))

    n_obs = len(Aij)
    d_x_zij = np.zeros(n_obs)
    d_zij_zij = np.zeros((n_obs,n_obs))
    d_zij_b0 = np.zeros(n_obs)
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries+j
                aij = Aij[ij]
                wij = Wij[ij]
                zij = mu[ij]
                xij = x[i]*x[j]
                pij = xij *zij/(1+zij*(b0+zij))
                den_1 = 1. + b0*zij + xij*zij
                den_2 = den_1 * (1+b0*zij)
                
                hess[i,i] += - pij*(1-pij)
                hess[i,j] = - pij*(1-pij)
                hess[j,i] = - pij*(1-pij)
                
                hess[i,-1] += - xij*zij**2/den_1**2
                hess[-1,i] += - xij*zij**2/den_1**2
                d_x_zij[ij] = xij*zij/den_1**2

                diffin = (b0+xij)*(1.+b0*zij) + b0*(1.+b0*zij+xij*zij)
                d_zij_zij[ij,ij] = -wij/zij - xij*zij**2/den_2**2 + xij*zij**2 * diffin/den_2**2
                d_zij_b0[ij] = xij*zij**2 *(2.*zij*(1+b0*zij)+xij*zij**2)/den_2**2

                hess[-1,-1] += - xij*zij**2 * (2.*zij*(1.+b0*zij) + xij*zij**2)/den_2**2
    for i in range(n_countries):            
        hess[i,n_countries:-1] = exogenous_variables.T @ d_x_zij
        hess[n_countries:-1,i] = exogenous_variables.T @ d_x_zij
    hess[n_countries:-1,n_countries:-1] = exogenous_variables.T @ d_zij_zij @ exogenous_variables
    hess[n_countries:-1,-1] = exogenous_variables.T @ d_zij_b0
    hess[-1,n_countries:-1] = exogenous_variables.T @ d_zij_b0

    return - hess


@jit(nopython=True,fastmath=True)
def hess_kIExp_topological_undirected(theta_0,Wij,exogenous_variables,params_exog,b0):
    """Computes the opposite of the topological hessian for the k-constrained Integrated Exponential model for Undirected networks"""
    
    
    n_countries = int(np.sqrt(len(Wij)))
    x = np.exp(-theta_0)
    b0 = b0[0]
    
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    hess = np.zeros((len(theta_0),len(theta_0)))

    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries+j
                zij = mu[ij]
                xij = x[i]*x[j]
                pij = xij *zij/(1+zij*(b0+zij))
                
                hess[i,i] += - pij*(1-pij)
                hess[i,j] = - pij*(1-pij)
                hess[j,i] = - pij*(1-pij)
                
    return - hess


@jit(nopython=True,fastmath=True)
def ll_kIExp_directed(params,Wij,exogenous_variables):
    """Computes the opposite of the log-likelihood for the k-constrained Integrated Exponential model for Directed networks"""
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[2*n_countries:-1]
    theta_out = params[:n_countries]
    theta_in = params[n_countries:2*n_countries]
    
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)
    
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                xij = x_out[i]*x_in[j]
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]

                bij = beta_0 + 1./zi
                ll += ai*np.log(xij) + np.log(bij) - bij*yi - np.log(xij+bij)
                
                
    return - ll 



@jit(nopython=True,fastmath=True)
def ll_kIExp_binary_directed(params,Wij,exogenous_variables):
    """Computes the opposite of the binary log-likelihood for the k-constrained Integrated Exponential model for Directed networks"""
    
    Aij = binarize(Wij)
    ll = 0
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[2*n_countries:-1]
    theta_out = params[:n_countries]
    theta_in = params[n_countries:2*n_countries]
    
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                xij = x_out[i]*x_in[j]
                yi = Wij[ij]
                ai = Aij[ij]
                zi = mu[ij]
                ll += ai*np.log(xij*zi) + (1.-ai)*np.log(1.+beta_0*zi) - np.log(1.+beta_0*zi + xij*zi) 
                

    return - ll 


@jit(nopython=True,fastmath=True)
def jac_kIExp_directed(params,Wij,exogenous_variables):
    """Computes the opposite of the jacobian for the k-constrained Integrated Exponential model for Directed networks"""
    
    Aij = binarize(Wij)
    n_obs = len(Aij)
    n_countries = int(np.sqrt(len(Aij)))
    
    n_exog_variables = exogenous_variables.shape[1]
    params_exog = params[2*n_countries:-1]
    theta_out = params[:n_countries]
    theta_in = params[n_countries:2*n_countries]
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)
    
    beta_0 = params[-1]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    jac = np.zeros(len(params))

    mu_matrix = np.zeros(len(Aij))
    
    adder_last = 0
    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                ji = j*n_countries + i
                
                yi = Wij[ij]
                aij = Aij[ij]
                aji = Aij[ji]
                zij = mu[ij]
                zji = mu[ji]
                
                xij = x_out[i]*x_in[j]
                xji = x_in[i]*x_out[j]
                
                pij = xij * zij/(1.+ beta_0*zij + xij*zij)
                pji = xji * zji/(1.+ beta_0*zji + xji*zji)
                
                yi_hat = pij*zij/(1+beta_0*zij)

                jac[i] += - aij + pij
                jac[i+n_countries] += - aji + pji
                adder_last += yi_hat - yi
                mu_matrix[ij] = (yi-yi_hat)/(zij)
                
    jac[2*n_countries:-1] = exogenous_variables.T @ mu_matrix
    jac[-1] = adder_last 
    
    return - jac

@jit(nopython=True,fastmath=True)
def jac_kIExp_topological_directed(theta,Wij,exogenous_variables,params_exog,beta_0):
    """Computes the opposite of the topological jacobian for the k-constrained Integrated Exponential model for Directed networks"""
    
    Aij = binarize(Wij)
    n_countries = int(np.sqrt(len(Aij)))
    
    theta_out = theta[:n_countries]
    theta_in = theta[n_countries:2*n_countries]
    x_out = np.exp(-theta_out)
    x_in = np.exp(-theta_in)
    beta_0 = beta_0[0]
    x1_beta1 = exogenous_variables @ params_exog

    mu = np.exp(x1_beta1)
    
    jac = np.zeros(len(theta))

    for i in range(n_countries):
        for j in range(n_countries):
            if j!=i:
                ij = i*n_countries + j
                ji = j*n_countries + i
                yi = Wij[ij]
                aij = Aij[ij]
                aji = Aij[ji]
                zij = mu[ij]
                zji = mu[ji]
                
                xij = x_out[i]*x_in[j]
                xji = x_in[i]*x_out[j]
                
                pij = xij * zij/(1+ beta_0*zij + xij*zij)
                pji = xji * zji/(1+ beta_0*zji + xji*zji)
                
                jac[i] += - aij + pij
                jac[i+n_countries] += - aji + pji
    
    return - jac





from .ll_Functions_conditional import *
from .ll_Functions_integrated import *
from .ll_Functions_econometrics import *
from .ensemble_Functions import *
from .netstats_Functions import *
from .solver_Functions import *
from .Undirected_class import *
from .Directed_class import *

__version__ = "0.0.1."
__author__ = """Marzio Di Vece (marzio.divece@imtlucca.it)"""


import numpy as np
from numba.typed import List
from . import ll_Functions_conditional as lfC
from . import ll_Functions_integrated as lfI
from . import ll_Functions_econometrics as lfE
from . import solver_Functions as sF
from . import netstats_Functions as nF
from . import ensemble_Functions as eF

class UndirectedGraph:
    """Undirected Graph instance must be initialised with adjacency weighted matrix
    On initialization it computes also the core network statistics and the binary adjacency matrix.
    
    :param adjacency: Weighted adjacency matrix 
    :type adjacency: np.ndarray, list
    
    """
    
    def __init__(
        self,
        adjacency=None,
    ):
        if adjacency is not None:
            if isinstance(adjacency,(np.ndarray,)):
                
                if len(adjacency.shape) == 2:
                    adjacency = nF.flatten(adjacency)
                elif len(adjacency.shape) > 2:
                    raise TypeError("Adjacency must be a 1-D or 2-D numpy array")
                
                        
            if not isinstance(adjacency,(np.ndarray,list)):
                raise TypeError("Adjacency must be a 1-D or 2-D numpy array")
            elif adjacency.size > 0:
                if np.sum(adjacency < 0):
                    raise TypeError(
                        "The adjacency matrix entries must be semi-definite positive."
                    )
                if isinstance(
                    adjacency, list
                ):
                    self.adjacency = np.array(adjacency)
                elif isinstance(adjacency, np.ndarray):
                    self.adjacency = adjacency
        else:
            raise TypeError("Adjacency matrix must be initialized.")
        
        if nF.is_symmetric(adjacency) == False:
            adjacency = nF.symmetrize(adjacency)
            raise TypeError("adjacency matrix has been symmetrized. If user does not want it symmetric please use DirectedGraph class!")
        self.binary_adjacency = nF.binarize(adjacency)
        
        self.degree = nF.deg(self.binary_adjacency)
        self.annd = nF.knn_single(self.binary_adjacency)
        self.clust = nF.clust_single_fast(self.binary_adjacency)
        self.strength = nF.strength(adjacency)
        self.anns = nF.stnn_single(adjacency,self.binary_adjacency)
        self.clust_w = nF.clust_w_single_fast(adjacency,self.binary_adjacency)
        self.n_edges = self.binary_adjacency.sum()
        self.n_nodes = int(np.sqrt(len(adjacency)))
        
        
        
        self.implemented_models = ["POIS","ZIP","NB2","ZINB","L-CGeom","k-CGeom","L-IGeom","k-IGeom",
                        "k-CExp","L-CExp","L-IExp","k-IExp","L-CPareto","k-CPareto","L-CGamma","k-CGamma","L-CLognormal","k-CLognormal"]
        self.discrete_models = ["POIS","ZIP","NB2","ZINB","L-IGeom","k-IGeom","L-CGeom","k-CGeom"]
        self.continuous_models = ["L-IExp","k-IExp","k-CExp","L-CExp","L-CPareto","k-CPareto","L-CGamma","k-CGamma","L-CLognormal","k-CLognormal"]
        self.implemented_network_statistics = ["degree","annd","clust","strength","anns","clust_w"]
        self.implemented_classifier_statistics = ["TPR","SPC","PPV","ACC","BACC","F1_score"]
        
        # Model Matrices
        self.model_weighted_adjacency = None
        self.model_binary_adjacency = None

        # Problem solutions
        self.params = None
        self.model = None
        self.optimization_time = None
        
        # Problem (reduced) residuals
        self.ll = None
        self.ll_binary = None
        self.jacobian = None
        self.norm = None
        self.aic = None
        self.aic_binary = None
        self.bic = None
        self.export_name = None

        
        # function
        self.args = None

        #classifier ensembler
        self.avg_nlinks = None
        self.std_nlinks = None
        self.avg_TPR = None
        self.std_TPR = None
        self.avg_SPC = None
        self.std_SPC = None
        self.avg_PPV = None
        self.std_PPV = None
        self.avg_ACC = None
        self.std_ACC = None
        self.avg_BACC = None
        self.std_BACC = None
        self.avg_F1score = None
        self.std_F1score = None
        

        #netstats ensembler
        self.avg_degree = None
        self.avg_annd = None
        self.avg_clust = None
        self.avg_strength = None
        self.avg_anns = None
        self.avg_clust_w = None
        
        self.std_degree = None
        self.std_annd = None
        self.std_clust = None
        self.std_strength = None
        self.std_anns = None
        self.std_clust_w = None

        #netstats_direct
        self.model_degree = None
        self.model_annd = None
        self.model_clust = None
        self.model_strength = None
        self.model_anns = None
        self.model_clust_w = None
        
        self.adjacency = adjacency  

        #weighted_prediction
        self.cond_w_score = None          
                
                     
        
    
    def solver(
        self,
        model,
        exogenous_variables = np.array([]),
        selection_variables = np.array([]),
        fixed_selection_params = np.array([]),
        verbose = False,
        print_steps = np.inf,
        imported_params = np.array([]),
        use_guess = np.array([]),
        maxiter = 10,
        tol = 1e-5,
    ):
        """Optimize chosen model for Undirected Graphs and compute model selection measures such as AIC and BIC. 
         
        :param model: Chosen model
        :type model: string
        
        :param exogenous_variables: Exogenous Variables used for the weighted step, in numpy 2-D format of the type N X k where N are the observations and k are the params,
        :type exogenous_variables: np.ndarray 
        :param selection_variables: Exogenous Variables used for the topological step, in numpy 2-D format of the type N X k where N are the observations and k are the params,
        :type selection_variables: np.ndarray 
        
        :param fixed_selection_params: Optional Constant Parameters for the topological step. When initialized different from null vector, it fixes the last "n" topological parameters, 
                                    when selection_variables are used and n is the size of fixed_selection_params 
        :type fixed_selection_params: np.ndarray 
        
        :param verbose: True if you want to see every n*print_steps iterations, default is False
        :type verbose: boolean
        
        :param print_steps: If verbose is True, you print on screen every n*print_steps iterations. default is 1
        :type print_steps: int
        
        :param imported_params: If used, uses wanted params as solution.
        :type imported_params: np.ndarray 
        
        :param use_guess: If used, uses wanted params as starters in the optimization.
        :type use_guess: np.ndarray 
        
        :param maxiter: Maximum Iterations of solver function.
        :type maxiter: int
        
        :param tol: tolerance for infinite norm in the optimization process
        :type tol: float 
       
        """
        
        self.model = model
        self.exogenous_variables = exogenous_variables
        self.selection_variables = selection_variables
        self.fixed_selection_params = fixed_selection_params
        # print('solver pre',len(self.fixed_selection_params))
        n_countries = int(np.sqrt(len(self.adjacency)))
            
        if model in ["ZIP", "ZINB"]:
            self.args = (self.adjacency,self.selection_variables,self.exogenous_variables,self.fixed_selection_params)
        elif model in ["L-CGeom", "L-CExp", "L-CPareto",  "L-CGamma", "L-CLognormal"]:
            self.args = (self.adjacency,self.selection_variables,self.exogenous_variables,self.fixed_selection_params)
        elif model == "Logit":
            self.args =   (self.adjacency,self.selection_variables,self.fixed_selection_params)     
        elif model in ["UBCM","DBCM"]:
            self.args = (self.adjacency,)
        else:
            self.args = (self.adjacency,self.exogenous_variables)
        
        self.conditional_models = ["L-CGeom","k-CGeom","L-CExp","k-CExp","L-CPareto","k-CPareto","L-CGamma","k-CGamma","L-CLognormal","k-CLognormal"]
        
        ll_fun = {
            "Logit": lambda x: -lfC.ll_logit(x, self.adjacency,self.selection_variables,self.fixed_selection_params),
            "UBCM": lambda x: -lfC.ll_BCM(
                x, self.adjacency),
            "DBCM": lambda x: - lfC.ll_DBCM(
                x, self.adjacency),
            "POIS": lambda x: -lfE.ll_POIS(x, *self.args),
            "ZIP": lambda x: - lfE.ll_ZIP(
                x, *self.args),
            "NB2": lambda x: -lfE.ll_NB2(x, *self.args),
            "ZINB": lambda x: - lfE.ll_ZINB(
                x, *self.args),
            "L-IGeom": lambda x: -lfI.ll_LIGeom(x, *self.args),
            "k-IGeom": lambda x: -lfI.ll_kIGeom_undirected(
                x, *self.args),
            "L-IExp": lambda x: -lfI.ll_LIExp(x, *self.args),
            "k-IExp": lambda x: -lfI.ll_kIExp_undirected(
                x, *self.args),
            "CGeom": lambda x: -lfC.ll_CGeom(x, self.adjacency,self.exogenous_variables),
            "CExp": lambda x: -lfC.ll_CExp(x, self.adjacency,self.exogenous_variables),
            "CPareto": lambda x: -lfC.ll_CPareto(x, self.adjacency,self.exogenous_variables),
            "CGamma": lambda x: -lfC.ll_CGamma(x, self.adjacency,self.exogenous_variables),
            "CLognormal": lambda x: -lfC.ll_CLognormal(x, self.adjacency,self.exogenous_variables), 
               
        }
        
        ll_fun_binary = {
            "Logit": lambda x: -lfC.ll_logit(x, self.adjacency,self.selection_variables,self.fixed_selection_params),
            "UBCM": lambda x: -lfC.ll_BCM(
                x, self.adjacency),
            "DBCM": lambda x: -lfC.ll_DBCM(
                x, self.adjacency),
            "POIS": lambda x: -lfE.ll_POIS_binary(x, *self.args),
            "ZIP": lambda x: -lfE.ll_ZIP_binary(
                x, *self.args),
            "NB2": lambda x: -lfE.ll_NB2_binary(x, *self.args),
            "ZINB": lambda x: -lfE.ll_ZINB_binary(
                x, *self.args),
            "L-IGeom": lambda x: -lfI.ll_LIGeom_binary(x, *self.args),
            "k-IGeom": lambda x: -lfI.ll_kIGeom_binary_undirected(
                x, *self.args),
            "L-IExp": lambda x: -lfI.ll_LIExp_binary(x, *self.args),
            "k-IExp": lambda x: -lfI.ll_kIExp_binary_undirected(
                x, *self.args),
                
        }
        
        jac_fun = {
            "Logit": lambda x: -lfC.jac_logit(x, self.adjacency,self.selection_variables,self.fixed_selection_params),
            "UBCM": lambda x: -lfC.jac_BCM(
                x, self.adjacency),
            "DBCM": lambda x: -lfC.jac_DBCM(
                x, self.adjacency),
            "POIS": lambda x: -lfE.jac_POIS(x, *self.args),
            "ZIP": lambda x: -lfE.jac_ZIP(
                x, *self.args),
            "NB2": lambda x: -lfE.jac_NB2(x, *self.args),
            "ZINB": lambda x: -lfE.jac_ZINB(
                x, *self.args),
            "L-IGeom": lambda x: -lfI.jac_LIGeom(x, *self.args),
            "k-IGeom": lambda x: -lfI.jac_kIGeom_undirected(
                x, *self.args),
            "L-IExp": lambda x: -lfI.jac_LIExp(x, *self.args),
            "k-IExp": lambda x: -lfI.jac_kIExp_undirected(
                x, *self.args),
            "CGeom": lambda x: -lfC.jac_CGeom(x, self.adjacency,self.exogenous_variables),
            "CExp": lambda x: -lfC.jac_CExp(x, self.adjacency,self.exogenous_variables),
            "CPareto": lambda x: -lfC.jac_CPareto(x, self.adjacency,self.exogenous_variables),
            "CGamma": lambda x: -lfC.jac_CGamma(x, self.adjacency,self.exogenous_variables),
            "CLognormal": lambda x: -lfC.jac_CLognormal(x, self.adjacency,self.exogenous_variables),    
        }
        
        
        if model in self.conditional_models:
            if model[:3] == "L-C":
                if len(imported_params) ==0:
                    solution = sF.solver(model=self.model,Wij=self.adjacency,selection_variables=self.selection_variables,exogenous_variables=self.exogenous_variables,
                                         fixed_selection_params=self.fixed_selection_params,tol=tol,use_guess= use_guess, 
                                         verbose=verbose, print_steps= print_steps, maxiter=maxiter )
                    self.params = solution
                else:
                    self.params = imported_params  
                n_selection_params = self.selection_variables.shape[1] - len(self.fixed_selection_params)

                ll_topological = ll_fun["Logit"](self.params[:n_selection_params])
                #lfC.ll_logit(self.params[:n_selection_params],self.adjacency,self.selection_variables,self.fixed_selection_params)
                ll_weighted = ll_fun[self.model[2:]](self.params[n_selection_params:])
                
                self.ll = ll_topological + ll_weighted
                self.ll_binary = ll_topological
                jac_topological = jac_fun["Logit"](self.params[:n_selection_params])
                jac_weighted = jac_fun[self.model[2:]](self.params[n_selection_params:])
                self.jacobian = np.concatenate((jac_topological,jac_weighted))
                self.norm = np.linalg.norm(self.jacobian,ord=np.inf)
                self.aic = nF.AIC(self.ll,len(self.params))
                self.aic_binary = nF.AIC(self.ll_binary,len(self.params[:n_selection_params]))
                self.bic = nF.BIC(self.ll,len(self.params),len(self.adjacency))
            
            elif model[:3] == "k-C":
                if len(imported_params) == 0:
                    self.input_model = self.model
                    self.input_model = self.input_model + "_undirected"
                    
                    solution = sF.solver(model=self.input_model,Wij=self.adjacency,selection_variables=self.selection_variables,exogenous_variables=self.exogenous_variables,
                                         fixed_selection_params=self.fixed_selection_params,tol=tol,use_guess= use_guess, 
                                         verbose=verbose, print_steps= print_steps, maxiter=maxiter )
                    self.params = solution
                else:
                    self.params = imported_params
                n_countries = int(np.sqrt(len(self.adjacency)))
                ll_topological = ll_fun["UBCM"](self.params[:n_countries])
                #lfC.ll_logit(self.params[:n_countries],self.adjacency,self.selection_variables,self.fixed_selection_params)
                ll_weighted = ll_fun[self.model[2:]](self.params[n_countries:])
                
                self.ll = ll_topological + ll_weighted
                self.ll_binary = ll_topological
                jac_topological = jac_fun["UBCM"](self.params[:n_countries])
                jac_weighted = jac_fun[self.model[2:]](self.params[n_countries:])
                self.jacobian = np.concatenate((jac_topological,jac_weighted))
                self.norm = np.linalg.norm(self.jacobian,ord=np.inf)
                self.aic = nF.AIC(self.ll,len(self.params))
                self.aic_binary = nF.AIC(self.ll_binary,len(self.params[:n_countries]))
                self.bic = nF.BIC(self.ll,len(self.params),len(self.adjacency))
                
        else:
                    
            if len(imported_params) == 0:
                self.input_model = self.model
                if self.input_model in ["k-IExp","k-IGeom"]:
                    self.input_model = self.input_model + "_undirected"
                solution = sF.solver(model=self.input_model,Wij=self.adjacency,selection_variables=self.selection_variables,exogenous_variables=self.exogenous_variables,
                                         fixed_selection_params=self.fixed_selection_params,tol=tol,use_guess= use_guess, 
                                         verbose=verbose, print_steps= print_steps, maxiter=maxiter )
                self.params = solution
            else:
                self.params = imported_params
                        
            self.ll = ll_fun[self.model](self.params)
            self.ll_binary = ll_fun_binary[self.model](self.params)
            self.jacobian = jac_fun[self.model](self.params)
            self.norm = np.linalg.norm(self.jacobian,ord=np.inf)
            self.aic = nF.AIC(self.ll,len(self.params))
            self.aic_binary = nF.AIC(self.ll_binary,len(self.params))
            self.bic = nF.BIC(self.ll,len(self.params),len(self.adjacency))

            
    

    def gen_ensemble(self,n_ensemble = 1000):
        """Generate an ensemble of -n_ensemble- networks according to the selected model
        
        :param n_ensemble: Number of wanted Graph realizations
        :type n_ensemble: int
        
        :return: N_obs X n_ensemble numpy matrix that collects all the ensemble adjacency matrices
        :rtype: np.ndarray
        """
        if self.model in self.continuous_models:
            w_mat_ensemble = eF.faster_ensemble_matrix_undirected(params=self.params,Wij=self.adjacency,
                                                        model=self.model,
                                                        exogenous_variables=self.exogenous_variables,
                                                        selection_variables=self.selection_variables,
                                                        fixed_selection_params = self.fixed_selection_params,
                                                        n_ensemble=n_ensemble)
        elif self.model in self.discrete_models:
            w_mat_ensemble = eF.discrete_ensemble_matrix_undirected(self.params,self.adjacency,self.model,self.exogenous_variables,
                                                                    self.selection_variables,self.fixed_selection_params,n_ensemble)
        else:
            raise TypeError("model not implemented!")        
        self.w_ensemble_matrix = w_mat_ensemble
        
    def reproduction_accuracy_s(self,percentiles = (2.5,97.5),stats=[]):
        """Computes RA_s, the percentage of nodes for which the network statistics are compatible with the graph ensemble according to a percentile confidence interval.
        To be used after .solve() and .gen_ensemble(). 
        
        :param percentiles: Explicit the percentiles used for the construction of the confidence interval, default is (2.5,97.5) for a 95 CI.
        :type percentiles: tuple
        :param stats: numpy array or list of network statistics for which it is possible to recover RA_s. The wanted stats must be in the list -.implemented_network_statistics-
        :type stats: list of strings
        
        :return: L-list where L is the number of statistics, the order follows the input stats array
        :rtype: list of float
        """
        
        if len(stats) == 0:
            stats = List(self.implemented_network_statistics)
        else:
            stats = List(stats)
        
        self.RA_s = nF.ensemble_coverage(self.w_ensemble_matrix,self.adjacency,percentiles,stats)
    
    def classification_measures(self,n_ensemble = 1000, percentiles = (2.5,97.5),stats=[]):
        """Computes Measures for the quality of classification, listed in self.implemented_classifier_statistics.
        To be used after .solve()
        :param n_ensemble: Wanted number of ensemble graphs, default is 1000.
        :type n_ensemble: int
        
        :param percentiles: Explicit the percentiles used for the construction of the confidence interval, default is (2.5,97.5) for a 95 CI.
        :type percentiles: tuple
        :param stats: numpy array or list of classifier statistics in string format.
        :type stats: list of strings
        
        """
        
        if len(stats) == 0:
            stats = self.implemented_classifier_statistics
        
        
        top_mat = nF.pij_matrix_undirected(self.params,self.model,self.adjacency,
        self.selection_variables,self.exogenous_variables,self.fixed_selection_params)
        self.model_binary_adjacency = top_mat

        if "TPR" in stats:
            self.avg_TPR, self.std_TPR, self.percentiles_TPR, self.array_TPR = nF.TPR_ensemble(top_mat,self.binary_adjacency,n_ensemble,percentiles)
        if "SPC" in stats:
            self.avg_SPC, self.std_SPC, self.percentiles_SPC, self.array_SPC = nF.SPC_ensemble(top_mat,self.binary_adjacency,n_ensemble,percentiles)
        if "PPV" in stats:
            self.avg_PPV, self.std_PPV, self.percentiles_PPV, self.array_PPV = nF.PPV_ensemble(top_mat,self.binary_adjacency,n_ensemble,percentiles)
        if "ACC" in stats:
            self.avg_ACC, self.std_ACC, self.percentiles_ACC, self.array_ACC = nF.ACC_ensemble(top_mat,self.binary_adjacency,n_ensemble,percentiles)
        if "BACC" in stats:
            self.avg_BACC, self.std_BACC, self.percentiles_BACC, self.array_BACC = nF.BACC_ensemble(top_mat,self.binary_adjacency,n_ensemble,percentiles)
        if "F1_score" in stats:
            self.avg_F1_score, self.std_F1_score, self.percentiles_F1_score, self.array_F1_score = nF.F1_score_ensemble(top_mat,self.binary_adjacency,n_ensemble,percentiles)
        
        
            
    def netstats_measures(self,percentiles = (2.5,97.5),stats=[] ):
        """Computes available network statistics, 
        To be used after .solve() and .gen_ensemble(). 
        
        :param percentiles: Explicit the percentiles used for the construction of the confidence interval, default is (2.5,97.5) for a 95 CI.
        :type percentiles: Tuple
        :param stats: numpy array or list of available network statistics. The wanted stats must be in the list -.implemented_network_statistics-
        :type stats: list
   
        """
        if len(stats) == 0:
            stats = self.implemented_network_statistics
        
        if "degree" in stats:
            self.avg_degree, self.std_degree, self.percentiles_degree, self.array_degree = nF.degree_ensemble(self.w_ensemble_matrix,percentiles)
        if "annd" in stats:
            self.avg_annd, self.std_annd, self.percentiles_annd, self.array_annd = nF.annd_ensemble(self.w_ensemble_matrix,percentiles)
        if "clust" in stats:
            self.avg_clust, self.std_clust, self.percentiles_clust, self.array_clust = nF.clust_ensemble(self.w_ensemble_matrix,percentiles)
        if "strength" in stats:
            self.avg_strength, self.std_strength, self.percentiles_strength, self.array_strength = nF.st_ensemble(self.w_ensemble_matrix,percentiles)
        if "anns" in stats:
            self.avg_anns, self.std_anns, self.percentiles_anns, self.array_anns = nF.anns_ensemble(self.w_ensemble_matrix,percentiles)
        if "clust_w" in stats:
            self.avg_clust_w, self.std_clust_w, self.percentiles_clust_w, self.array_clust_w = nF.cw_ensemble(self.w_ensemble_matrix,percentiles)
        
        
        
    def reproduction_accuracy_w(self,percentiles = (2.5,97.5)):
        """Computes RA_w, the percentage of couples for which the weights are compatible with the graph ensemble according to a percentile confidence interval.
        To be used after .solve() and .gen_ensemble(). 
        
        :param percentiles: Explicit the percentiles used for the construction of the confidence interval, default is (2.5,97.5) for a 95 CI.
        :type percentiles: Tuple
        
        :return: A number (0,1) that explicits RA_w
        :rtype: np.ndarray
        """
        self.RA_w = nF.weighted_coverage(self.w_ensemble_matrix,self.adjacency, percentiles)
        
